from enum import Enum

from pypika import functions as fn

from .substitution import Substitution


class AggregateType(Enum):
    unknown = "UNKNOWN"
    count = "count"
    sum = "sum"
    number = "number"


class Aggregate:
    def __init__(self, measure, table):
        sql = measure.settings["sql"]
        self.substitution = Substitution(sql, table)
        self.measure = measure
        self.table = table
        self.sql = self.substitution.sql
        self.aggregateType = AggregateType.unknown
        self.getAggregateType()

    def getAggregateType(self):
        type_ = self.measure.settings["type"]
        if type_ == AggregateType.sum.value:
            self.aggregateType = AggregateType.sum
            self.setAggregateSQLSum()
        elif type_ == AggregateType.count.value:
            self.aggregateType = AggregateType.count
            self.setAggregateSQLCount()
        elif type_ == AggregateType.number.value:
            self.aggregateType = AggregateType.number
            self.setAggregateSQLNumber()
        else:
            self.aggregateType = AggregateType.unknown
            raise Exception(f"Aggregate Type {type_} not implemented yet")

    def setAggregateSQLSum(self):
        self.sql = fn.Coalesce(fn.Sum(self.sql), 0, alias=self.substitution.alias)

    def setAggregateSQLCount(self):
        self.sql = fn.Coalesce(fn.Count(self.sql), 0, alias=self.substitution.alias)

    def setAggregateSQLNumber(self):
        self.sql.alias = self.substitution.alias



import pytest
import yaml
import os
import shutil

from meltano.core.project_add_service import ProjectAddService
from meltano.core.plugin_discovery_service import PluginNotFoundError
from meltano.core.plugin import PluginType


class TestProjectAddService:
    def test_default_init_should_not_fail(self, project):
        add_service = ProjectAddService(project)
        assert add_service

    def test_missing_plugin_exception(self, project):
        add_service = ProjectAddService(project)
        try:
            add_service.add(PluginType.EXTRACTORS, "tap-missing")
        except Exception as e:
            assert type(e) is PluginNotFoundError

    def test_add_extractor(self, project):
        add_service = ProjectAddService(project)
        added = add_service.add(PluginType.EXTRACTORS, "tap-first")

        assert added.canonical() in project.meltano[PluginType.EXTRACTORS]



import pytest

from meltano.core.project_init_service import ProjectInitService
from meltano.core.project_add_service import ProjectAddService
from meltano.core.config_service import ConfigService


PROJECT_NAME = "a_meltano_project"


@pytest.fixture
def project_init_service():
    return ProjectInitService()


@pytest.fixture
def project_add_service(project):
    return ProjectAddService(project)


@pytest.fixture
def config_service(project):
    return ConfigService(project)


@pytest.fixture(scope="class")
def project(test_dir):
    service = ProjectInitService(PROJECT_NAME)
    project = service.init()

    # cd into the new project root
    project.chdir()
    return project



import os
import logging
import pandas as pd

from sqlalchemy import create_engine, inspect, Table
from sqlalchemy.schema import CreateSchema

from snowflake.sqlalchemy import URL


class SnowflakeLoader:
    def __init__(self, **kwargs):
        """
        :param entity_name: str Name of the entity to load
        :param kwargs:
        """
        self.entity_name = kwargs["entity_name"]
        self.extractor = kwargs["extractor"]
        self.table = self.extractor.tables.get(self.entity_name)
        self.index_elements = self.extractor.primary_keys.get(self.entity_name)

        if "connection_string" in kwargs:
            self.engine = create_engine(kwargs["connection_string"])
        else:
            self.engine = create_engine(
                URL(
                    user=os.environ.get("SNOWFLAKE_USERNAME"),
                    password=os.environ.get("SNOWFLAKE_PASSWORD"),
                    account=os.environ.get("SNOWFLAKE_ACCOUNT_NAME"),
                    database=os.environ.get("SNOWFLAKE_DB_NAME"),
                    warehouse=os.environ.get("SNOWFLAKE_WAREHOUSE"),
                )
            )

    def schema_apply(self) -> None:
        inspector = inspect(self.engine)

        all_schema_names = inspector.get_schema_names()
        if not (self.table.schema in all_schema_names):
            logging.info(f"Schema {self.table.schema} does not exist -> creating it ")
            self.engine.execute(CreateSchema(self.table.schema))

        all_table_names = inspector.get_table_names(self.table.schema)
        if not (self.table.name in all_table_names):
            logging.info(f"Table {self.table.name} does not exist -> creating it ")
            self.table.metadata.create_all(self.engine)

    def load(self, df: pd.DataFrame) -> None:
        if not df.empty:
            # Convert tricky NaN and NaT values to None
            #  so that they are properly stored as NULL values
            df_to_load = df.astype(object).where(pd.notnull(df), None)

            logging.info(f"Loading data to Snowflake for {self.entity_name}")

            if len(self.table.primary_key) > 0:
                # We have to use Snowflake's Merge in order to Upsert

                # Create Temporary table to load the data to
                tmp_table = self.create_tmp_table()

                # Insert data to temporary table
                df_to_load.to_sql(
                    name=tmp_table.name,
                    con=self.engine,
                    schema=tmp_table.schema,
                    if_exists="append",
                    index=False,
                )

                # Merge Temporary Table into the Table we want to load data into
                merge_stmt = self.generate_merge_stmt(tmp_table.name)

                try:
                    connection = self.engine.connect()
                    connection.execute(merge_stmt)
                finally:
                    connection.close()

                # Drop the Temporary Table
                tmp_table.drop(self.engine)
            else:
                # Just Insert (append) as no conflicts can arise
                df_to_load.to_sql(
                    name=self.table.name,
                    con=self.engine,
                    schema=self.table.schema,
                    if_exists="append",
                    index=False,
                )
        else:
            logging.info(f"DataFrame {df} is empty -> skipping it")

    def create_tmp_table(self) -> Table:
        columns = [c.copy() for c in self.table.columns]
        tmp_table = Table(
            f"tmp_{self.table.name}",
            self.table.metadata,
            *columns,
            schema=self.table.schema,
            keep_existing=True,
        )

        tmp_table.drop(self.engine, checkfirst=True)
        tmp_table.create(self.engine)

        return tmp_table

    def generate_merge_stmt(self, tmp_table_name: str) -> str:
        """
        Generate a merge statement for Merging a temporary table into the
          main table.

        The Structure of Merge in Snowflake is as follows:
          MERGE INTO <target_table> USING <source_table>
            ON <join_expression>
          WHEN MATCHED THEN UPDATE SET <update_clause>
          WHEN NOT MATCHED THEN INSERT (source_atts) VALUES {insert_clause}

        In this simple form, it has the same logic as UPSERT in Postgres
            INSERT ... ... ...
            ON CONFLICT ({pkey}) DO UPDATE SET {update_clause}
        """
        target_table = f"{self.table.schema}.{self.table.name}"
        source_table = f"{self.table.schema}.{tmp_table_name}"

        # Join using all primary keys
        joins = []
        for primary_key in self.table.primary_key:
            pk = primary_key.name
            joins.append(f"{target_table}.{pk} = {source_table}.{pk}")
        join_expr = " AND ".join(joins)

        attributes_target = []
        attributes_source = []
        update_sub_clauses = []

        for column in self.table.columns:
            attr = column.name
            attributes_target.append(attr)
            attributes_source.append(f"{source_table}.{attr}")

            if attr not in self.table.primary_key:
                update_sub_clauses.append(f"{attr} = {source_table}.{attr}")

        update_clause = ", ".join(update_sub_clauses)
        matched_clause = f"WHEN MATCHED THEN UPDATE SET {update_clause}"

        source_attributes = ", ".join(attributes_target)
        target_attributes = ", ".join(attributes_source)

        insert_clause = f"({source_attributes}) VALUES ({target_attributes})"
        not_matched_clause = f"WHEN NOT MATCHED THEN INSERT {insert_clause}"

        merge_stmt = f"MERGE INTO {target_table} USING {source_table} ON {join_expr} {matched_clause} {not_matched_clause}"

        return merge_stmt

    def grant_privileges(self, role: str) -> None:
        # GRANT access to the created schema and tables to the provided role
        try:
            db_name = os.environ.get("SNOWFLAKE_DB_NAME")
            connection = self.engine.connect()
            grant_stmt = (
                f"GRANT USAGE ON SCHEMA {db_name}.{self.table.schema} TO ROLE {role};"
            )
            connection.execute(grant_stmt)
            grant_stmt = f"GRANT SELECT ON ALL TABLES IN SCHEMA {db_name}.{self.table.schema} TO ROLE {role};"
            connection.execute(grant_stmt)
        finally:
            connection.close()



import os
from dotenv import load_dotenv

load_dotenv()


def get_env_variable(name):
    try:
        return os.environ[name]
    except KeyError:
        message = f"Expected environment variable '{name}' not set."
        raise Exception(message)


# the values of those depend on your setup
MELTANO_POSTGRES_URL = get_env_variable("MELTANO_POSTGRES_URL")
MELTANO_POSTGRES_USER = get_env_variable("MELTANO_POSTGRES_USER")
MELTANO_POSTGRES_PASSWORD = get_env_variable("MELTANO_POSTGRES_PASSWORD")
MELTANO_POSTGRES_DB = get_env_variable("MELTANO_POSTGRES_DB")
LOG_PATH = get_env_variable("MELTANO_ANALYZE_LOG_PATH")
ENV = "development"

API_ROOT_DIR = os.path.abspath(os.path.dirname(__file__))
TEMP_FOLDER = os.path.join(API_ROOT_DIR, "static/tmp")
PROJECT_ROOT_DIR = os.path.dirname(API_ROOT_DIR)

SQLALCHEMY_DATABASE_URI = f"postgresql+psycopg2://{MELTANO_POSTGRES_USER}:{MELTANO_POSTGRES_PASSWORD}@{MELTANO_POSTGRES_URL}/{MELTANO_POSTGRES_DB}"

SQLALCHEMY_TRACK_MODIFICATIONS = False

# SQLALCHEMY_ECHO=True

THREADS_PER_PAGE = 2

SECRET_KEY = "damnitjanice"

meltano_model_path = get_env_variable("MELTANO_MODEL_DIR")
meltano_transform_path = get_env_variable("MELTANO_TRANSFORM_DIR")



import os
import yaml
import json
import click
from urllib.parse import urlparse
from . import cli
from meltano.core.project_add_service import (
    ProjectAddService,
    PluginNotSupportedException,
)
from meltano.core.plugin_install_service import PluginInstallService
from meltano.core.plugin_discovery_service import PluginNotFoundError
from meltano.core.plugin import PluginType
from meltano.core.database_add_service import DatabaseAddService
from meltano.core.project import Project


@cli.group()
def add():
    pass


@add.command()
@click.option("--name", prompt="Database connection name")
@click.option("--host", prompt="Database host")
@click.option("--database", prompt="Database database")
@click.option("--schema", prompt="Database schema")
@click.option("--username", prompt="Database username")
@click.option(
    "--password", prompt="Database password", hide_input=True, confirmation_prompt=True
)
def database(name, host, database, schema, username, password):
    project = Project.find()
    database_add_service = DatabaseAddService(project)
    database_add_service.add(
        name=name,
        host=host,
        database=database,
        schema=schema,
        username=username,
        password=password,
    )
    click.secho("database yml file updated", fg="green")


@add.command()
@click.argument("plugin_name")
def extractor(plugin_name):
    project = Project.find()
    add_plugin(project, PluginType.EXTRACTORS, plugin_name)


@add.command()
@click.argument("plugin_name")
def loader(plugin_name):
    project = Project.find()
    add_plugin(project, PluginType.LOADERS, plugin_name)


def add_plugin(project: Project, plugin_type: PluginType, plugin_name: str):
    try:
        add_service = ProjectAddService(project)
        plugin = add_service.add(plugin_type, plugin_name)
        click.secho(f"{plugin_name} added to your meltano.yml config", fg="green")

        install_service = PluginInstallService(project)
        click.secho("Activating your virtual environment", fg="green")
        run_venv = install_service.create_venv(plugin)

        if run_venv["stdout"]:
            click.echo(run_venv["stdout"])
        if run_venv["stderr"]:
            click.secho(run_venv["stderr"], fg="red")

        click.secho(f"Installing {plugin_name} via pip...", fg="green")
        run_install_plugin = install_service.install_plugin(plugin)
        if run_install_plugin["stdout"]:
            click.echo(run_install_plugin["stdout"])
        if run_install_plugin["stderr"]:
            click.secho(run_install_plugin["stderr"], fg="red")

        click.secho(f"Added and installed {plugin_type} {plugin_name}", fg="green")
    except PluginNotSupportedException:
        click.secho(f"The {plugin_type} {plugin_name} is not supported", fg="red")
        raise click.Abort()
    except PluginNotFoundError as e:
        click.secho(f"{plugin_type.title()} {plugin_name} not supported", fg="red")
        raise click.Abort()



import subprocess
from .project import Project


class VenvService:
    def __init__(self, project):
        self.project = project

    def create(self, namespace="", name=""):
        venv_path = self.project.venvs_dir(namespace, name)
        run_venv = subprocess.run(
            ["python3", "-m", "venv", venv_path],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            universal_newlines=True,
        )

        return {"stdout": run_venv.stdout, "stderr": run_venv.stderr}

    def install(self, pip_url, namespace="", name=""):
        pip_install_path = self.exec_path("pip", namespace=namespace, name=name)
        run_pip_install = subprocess.run(
            [pip_install_path, "install", *pip_url.split(" ")],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            universal_newlines=True,
        )

        return {"stdout": run_pip_install.stdout, "stderr": run_pip_install.stderr}

    def exec_path(self, bin, name=None, namespace=""):
        name = name or bin
        return self.project.venvs_dir(namespace, name).joinpath("bin", bin)



from enum import Enum

from .substitution import Substitution
from ..models.data import View, Dimension


class JoinType(Enum):
    left_join = "left"
    inner_join = "inner"
    full_outer_join = "full_outer"
    cross_join = "cross"


class Join:
    def __init__(self, join, view, table):
        self.table = table
        self.view = view
        join_type = self.get_join_type(join)
        join_sql = self.join_sql(join, table)
        self.sql = " ".join([join_type.value, join_sql])

    def get_join_type(self, join):
        if not "type" in join.settings:
            return JoinType.left_join
        else:
            type_of_join = join.settings["type"].lower()
            if type_of_join == JoinType.inner_join.value:
                return JoinType.inner_join
            elif type_of_join == JoinType.full_outer_join.value:
                return JoinType.full_outer_join
            elif type_of_join == JoinType.cross_join.value:
                return JoinType.cross_join

    def join_sql(self, join, table):
        sql = join.settings["sql_on"]
        results = Substitution.placeholder_match(sql)
        related_view = View.query.filter(View.name == join.name).first()

        def dimension_actual_name(result):
            (joined_view, joined_dimension) = result.split(".")
            queried_view = self.view
            if queried_view.name != joined_view:
                queried_view = related_view
            _dimensions = Dimension.query.filter(
                Dimension.name == joined_dimension
            ).all()
            queried_dimension = (
                Dimension.query.join(View, Dimension.view_id == View.id)
                .filter(Dimension.name == joined_dimension)
                .filter(View.id == queried_view.id)
                .first()
            )

            return queried_dimension.table_column_name

        inner_results = list(map(dimension_actual_name, results.inner))

        Substitution.variable_replace(results)

        for i, result in enumerate(results.outer):
            sql = sql.replace(result, inner_results[i])
        table_name = related_view.settings["sql_table_name"]
        return f"{table_name} AS {join.name} ON {sql}"



from .base import Base
from ..app import db


class Settings(Base):
    __tablename__ = "setting"
    settings = db.Column(db.JSON(), nullable=False, default={"connections": []})

    def serializable(self):
        return {"settings": self.settings}

    def __repr__(self):
        return f"<Settings for {self.project} id: {self.id}>"



import pytest
import os

from meltano.core.venv_service import VenvService


class TestVenvService:
    @pytest.fixture
    def subject(self, project):
        return VenvService(project)

    def test_create(self, project, subject):
        subject.create(name="name", namespace="namespace")
        venv_dir = subject.project.venvs_dir("namespace", "name")

        # ensure the venv is created
        assert venv_dir.exists()

        # ensure that the binary is python3
        assert os.path.samefile(
            venv_dir.joinpath("bin/python"), venv_dir.joinpath("bin/python3")
        )



import pytest
import yaml
import os
import shutil

from meltano.core.config_service import ConfigService
from meltano.core.plugin import Plugin, PluginType


def make_meltano_yml(project):
    with open(project.meltanofile, "w") as f:
        f.write(
            yaml.dump(
                {
                    "extractors": [
                        {
                            "name": "first",
                            "pip_url": "git+https://gitlab.com/meltano/tap-first.git",
                        }
                    ],
                    "loaders": [
                        {
                            "name": "csv",
                            "pip_url": "git+https://gitlab.com/meltano/target-csv.git",
                        }
                    ],
                }
            )
        )


def make_database_yml(project):
    with open(project.meltano_dir(".database_test.yml"), "w") as f:
        f.write(
            yaml.dump(
                {
                    "database": "test",
                    "host": "127.0.0.1",
                    "name": "test",
                    "password": "hatpants",
                    "root_name": "TEST",
                    "schema": "test",
                    "username": "test",
                }
            )
        )


class TestConfigService:
    @pytest.fixture
    def subject(self, project):
        make_meltano_yml(project)
        make_database_yml(project)
        return ConfigService(project)

    def test_default_init_should_not_fail(self, subject):
        assert subject

    def test_get_extractors(self, subject):
        extractors = list(subject.get_extractors())
        assert len(extractors) == 1
        assert all(map(lambda p: p.type == PluginType.EXTRACTORS, extractors))

    def test_get_loaders(self, subject):
        loaders = list(subject.get_loaders())
        assert len(loaders) == 1
        assert all(map(lambda p: p.type == PluginType.LOADERS, loaders))

    def test_database_config(self, subject):
        assert subject.get_database("test") == {
            "database": "test",
            "host": "127.0.0.1",
            "name": "test",
            "password": "hatpants",
            "root_name": "TEST",
            "schema": "test",
            "username": "test",
        }



import os
from unittest.mock import patch
from functools import partial
from click.testing import CliRunner

from meltano.cli import cli
from meltano.core.runner.singer import SingerRunner
from meltano.core.runner.dbt import DbtRunner
from meltano.core.dbt_service import DbtService


PERFORM_TEST_ARGS = [
    "elt",
    "test_job_id",
    "--extractor",
    "tap-test",
    "--loader",
    "target-test",
]


def test_elt(request, project):
    cli_runner = CliRunner()

    result = cli_runner.invoke(cli, ["elt"])
    assert result.exit_code == 2

    # exit cleanly when everything is fine
    with patch.object(SingerRunner, "perform", return_value=None), patch.object(
        DbtRunner, "perform", return_value=None
    ), patch.object(DbtService, "deps", return_value=None):
        result = cli_runner.invoke(cli, PERFORM_TEST_ARGS)
        assert result.exit_code == 0

    # aborts when there is an exception
    with patch.object(SingerRunner, "perform", side_effect=Exception):
        result = cli_runner.invoke(cli, PERFORM_TEST_ARGS)
        assert result.exit_code == 1

    with patch.object(SingerRunner, "perform", side_effect=Exception):
        result = cli_runner.invoke(cli, PERFORM_TEST_ARGS)
        assert result.exit_code == 1



import os
import logging
import click

from . import cli
from .params import db_options
from meltano.core.runner.singer import SingerRunner
from meltano.core.runner.dbt import DbtRunner
from meltano.core.dbt_service import DbtService
from meltano.core.project import Project
from meltano.core.plugin import PluginType


@cli.command()
@db_options
@click.argument("job_id", envvar="MELTANO_JOB_ID")
@click.option(
    "--extractor",
    help="Which extractor should be used in this extraction",
    required=True,
)
@click.option(
    "--loader", help="Which loader should be used in this extraction", required=True
)
@click.option("--dry", help="Do not actually run.", is_flag=True)
@click.option(
    "--transform", type=click.Choice(["skip", "only", "auto"]), default="auto"
)
def elt(job_id, extractor, loader, dry, transform):
    project = Project.find()

    singer_runner = SingerRunner(
        project,
        job_id=job_id,
        run_dir=os.getenv("SINGER_RUN_DIR", project.meltano_dir("run")),
        target_config_dir=project.meltano_dir(PluginType.LOADERS, loader),
        tap_config_dir=project.meltano_dir(PluginType.EXTRACTORS, extractor),
    )

    dbt_runner = DbtRunner(project)

    try:
        if transform != "only":
            click.echo("Running extract & load...")
            singer_runner.perform(extractor, loader, dry_run=dry)
            click.secho("Extract & load complete!", fg="green")
        else:
            click.secho("Extract & load skipped.", fg="yellow")

        if transform != "skip":
            click.echo("Running transformation...")
            dbt_runner.perform(dry_run=dry)
            click.secho("Transformation complete!", fg="green")
        else:
            click.secho("Transformation skipped.", fg="yellow")
    except Exception as err:
        logging.exception(err)
        click.secho(f"Extraction failed: {err}.", fg="red")
        raise click.Abort()



import os
import yaml
import click

from .project import Project
from .venv_service import VenvService
from .dbt_service import DbtService


class ProjectInitServiceError(Exception):
    pass


class ProjectInitService:
    def __init__(self, project_name):
        self.initialize_file = os.path.join(os.path.dirname(__file__), "initialize.yml")
        self.project_name = project_name.lower()

    def init(self) -> Project:
        default_project_yaml = yaml.load(open(self.initialize_file))
        try:
            os.mkdir(self.project_name)
        except Exception as e:
            raise ProjectInitServiceError

        new_project = Project(self.project_name)
        self.project_echo("", True)
        for name in default_project_yaml.keys():
            if name.startswith("/"):
                self.create_dir(name)
            else:
                self.create_file(name, default_project_yaml[name])

        return new_project

    def create_file(self, name, content):
        with open(self.join_with_project_base(name), "w") as f:
            f.write(content)
            self.project_echo(name, False, True)

    def create_dir(self, name):
        current_dir = name[1:]
        os.mkdir(os.path.join(".", self.project_name, current_dir))
        self.project_echo(current_dir, False, True)

    def project_echo(self, filename="", star=False, check=False):
        star = "⭐" if star else ""
        check = "✅" if check else ""
        click.secho(f"{star}{check}\tCreated", fg="blue", nl=False)
        click.echo(f" ./{self.project_name}/{filename}")

    def echo_instructions(self):
        click.secho(f"🚀\t{self.project_name}", fg="green", nl=False)
        click.echo(" has been created. Next steps:")
        click.echo(f"🚪\tcd ", nl=False)
        click.secho(self.project_name, fg="green")
        click.echo(f"✏️\tEdit", nl=False)
        click.secho(" .env", fg="green")
        click.echo(f"🏃\tRun", nl=False)
        click.secho(" source .env", fg="green")
        click.echo("Try it out:")
        click.secho("🏃\tRun", nl=False)
        click.secho(" meltano add extractor tap-first", fg="green")
        click.secho("🏃\tRun", nl=False)
        click.secho(" meltano add loader target-postgres", fg="green")
        click.secho("🏃\tRun", nl=False)
        click.secho(
            " meltano elt job_id1 --extractor tap-first --loader target-postgres",
            fg="green",
        )
        click.echo("WOW! NEATO!")
        click.echo("📖\tRead the Meltano README.", nl=False)
        click.secho(
            " https://gitlab.com/meltano/meltano/blob/master/README.md", fg="red"
        )
        click.echo("✏️\tEdit the meltano.yml file.")

    def join_with_project_base(self, filename):
        return os.path.join(".", self.project_name, filename)

    def install_dbt(self, project: Project):
        venv_service = VenvService(project)
        venv_service.create(name="dbt")
        venv_service.install(name="dbt", pip_url="dbt")



import click
from . import cli
from meltano.core.plugin_install_service import PluginInstallService
from meltano.core.plugin_discovery_service import PluginNotFoundError
from meltano.core.project import Project
from meltano.core.project_add_service import ProjectAddService


def install_status_update(data):
    plugin = data["plugin"]

    if data["status"] == "running":
        click.secho(f"Installing {plugin.type}: {plugin.name}", fg="yellow")
    if data["status"] == "error":
        click.secho(f"An error occured: {data['message']}.", fg="red")
    if data["status"] == "warning":
        click.secho(f"Warning! {data['message']}.", fg="yellow")
    if data["status"] == "success":
        click.secho(
            f"{plugin.type.title()}: {plugin.name} installed successfully", fg="green"
        )


@cli.command()
def install():
    project = Project.find()
    install_service = PluginInstallService(project)
    install_status = install_service.install_all_plugins(install_status_update)
    num_installed = len(install_status["installed"])
    click.secho(f"{num_installed} plugins installed successfully", fg="green")



import datetime
import json
import os

import aiohttp
import pandas as pd
from dateutil.relativedelta import relativedelta
from meltano.core.extract_utils import fetch_urls
from pandas.io.json import json_normalize
from sqlalchemy import Table, Column, String, MetaData, TIMESTAMP

REQUEST_TIMEOUT_SEC = 30
CONNECTIONS_LIMIT = 100
FASTLY_API_SERVER = "https://api.fastly.com/"
FASTLY_HEADERS = {
    "Fastly-Key": os.getenv("FASTLY_API_TOKEN"),
    "Accept": "application/json",
}

fastly_metadata = MetaData()
line_items = Table(
    "line_items",  # Name of the table
    fastly_metadata,
    Column("id", String, primary_key=True),
    Column("aria_invoice_id", String),
    Column("line_number", String),
    Column("description", String),
    Column("service_no", String),
    Column("service_name", String),
    Column("units", String),
    Column("rate_per_unit", String),
    Column("amount", String),
    Column("rate_schedule_no", String),
    Column("rate_schedule_tier_no", String),
    Column("usage_type_no", String),
    Column("usage_type_cd", String),
    Column("plan_no", String),
    Column("plan_name", String),
    Column("client_service_id", String),
    Column("credit_coupon_code", String, nullable=True),
    Column("created_at", TIMESTAMP),
    Column("updated_at", TIMESTAMP),
    Column("deleted_at", TIMESTAMP, nullable=True),
    # Column from the meta of the response
    Column("customer_id", String),
    Column("invoice_id", String),
    Column("start_time", TIMESTAMP, nullable=True),
    Column("end_time", TIMESTAMP, nullable=True),
    schema="fastly",
)


class FastlyExtractor:
    """
    Extractor for the Fastly Billing API
    """

    def __init__(self):
        self.name = "fastly"
        today = datetime.date.today()
        self.this_month = datetime.date(year=today.year, month=today.month, day=1)
        self.start_date = datetime.date(
            2017, 8, 1
        )  # after this period billing data starts
        self.entities: list = ["line_items"]
        self.tables = {"line_items": line_items}
        self.primary_keys = {"line_items": ["id"]}

    def get_billing_urls(self):
        urls = []
        date = self.start_date
        while date < self.this_month:
            billing_endpoint = f"billing/v2/year/{date.year:04}/month/{date.month:02}"
            urls.append(f"{FASTLY_API_SERVER}{billing_endpoint}")
            date += relativedelta(months=1)
        return urls

    def extract(self, entity_name):
        if entity_name == "line_items":
            timeout = aiohttp.ClientTimeout(total=REQUEST_TIMEOUT_SEC)
            responses = fetch_urls(
                urls=self.get_billing_urls(), headers=FASTLY_HEADERS, timeout=timeout
            )
            for resp in responses:
                resp_json = json.loads(resp)
                df = json_normalize(
                    resp_json,
                    record_path="line_items",
                    # TODO: generate from the Table def
                    meta=["customer_id", "invoice_id", "end_time", "start_time"],
                )
                datetime_cols = [
                    "created_at",
                    "deleted_at",
                    "updated_at",
                    "end_time",
                    "start_time",
                ]
                # TODO: generate datetime cols from the Table def
                df[datetime_cols].apply(pd.to_datetime)
                yield df



import yaml

from meltano.core.project_add_service import ProjectAddService
from meltano.core.plugin_discovery_service import PluginDiscoveryService
from meltano.core.plugin import PluginType


def test_add_extractor(project):
    service = ProjectAddService(project)
    service.add(PluginType.EXTRACTORS, "tap-first")

    assert PluginType.EXTRACTORS in project.meltano
    assert any(
        plugin["name"] == "tap-first"
        for plugin in project.meltano[PluginType.EXTRACTORS]
    )


def test_add_loader(project):
    service = ProjectAddService(project)
    service.add(PluginType.LOADERS, "target-csv")

    assert PluginType.LOADERS in project.meltano
    assert any(
        plugin["name"] == "target-csv" for plugin in project.meltano[PluginType.LOADERS]
    )



import os
import yaml
import logging
from pathlib import Path
from typing import Union, Dict
from contextlib import contextmanager
from functools import wraps

from .plugin import Plugin


class Project:
    """
    Represent the current Meltano project from a file-system
    perspective.
    """

    _meltano = {}

    def __init__(self, root: Union[Path, str]):
        self.root = Path(root)

    def chdir(self):
        os.chdir(self.root)
        self.root = Path(".")

    @classmethod
    def find(self, from_dir: Union[Path, str] = ".", chdir=True):
        """
        Recursively search for a `meltano.yml` file. Once found,
        return a `Project` correct dir.
        """
        cwd = os.getcwd()

        while not Project(".").meltanofile.exists():
            if os.getcwd() == "/":
                raise Exception("Cannot go any further.")

            os.chdir("..")

        if not chdir:
            os.chdir(cwd)

        return Project(os.getcwd())

    @property
    def meltano(self) -> Dict:
        """Return a copy of the current meltano config"""
        if not self._meltano:
            self._meltano = yaml.load(self.meltanofile.open()) or {}
        return self._meltano.copy()

    @contextmanager
    def meltano_update(self):
        """
        Yield the current meltano configuration and update the meltanofile
        if the context ends gracefully.
        """
        try:
            meltano_update = self.meltano
            yield meltano_update

            # save it
            with self.meltanofile.open("w") as meltanofile:
                meltanofile.write(yaml.dump(meltano_update, default_flow_style=False))

            # update the cache
            self._meltano = meltano_update
        except Exception as err:
            logging.error(f"Could not update meltano.yml: {err}")
            raise

    def makedirs(func):
        @wraps(func)
        def decorate(*args, **kwargs):
            path = func(*args, **kwargs)

            # if there is an extension, only create the base dir
            _, ext = os.path.splitext(path)
            if ext:
                dir = os.path.dirname(path)
            else:
                dir = path

            os.makedirs(dir, exist_ok=True)
            return path

        return decorate

    @property
    def meltanofile(self):
        return self.root.joinpath("meltano.yml")

    @makedirs
    def meltano_dir(self, *joinpaths):
        return self.root.joinpath(".meltano", *joinpaths)

    @makedirs
    def venvs_dir(self, *prefixes):
        return self.meltano_dir(*prefixes, "venv")

    @makedirs
    def run_dir(self, *joinpaths):
        return self.meltano_dir("run", *joinpaths)

    @makedirs
    def plugin_dir(self, plugin: Plugin, *joinpaths):
        return self.meltano_dir(plugin.type, plugin.name, *joinpaths)



from . import Plugin
from meltano.core.error import PluginInstallWarning


class PluginMissingError(Exception):
    """
    Base exception when a plugin seems to be missing.
    """

    def __init__(self, plugin_or_name):
        if isinstance(plugin_or_name, Plugin):
            self.plugin_name = plugin_or_name.name
        else:
            self.plugin_name = plugin_or_name


class TapDiscoveryError(PluginInstallWarning):
    def __str__(self):
        return (
            "Running --discover on the tap failed, "
            "no catalog will be provided."
            "Some taps requires a valid catalog file "
            "with some selected streams."
        )



from typing import Iterator

import pandas as pd
from meltano.core.extract_utils import tables_from_manifest
from sqlalchemy import MetaData

sfdc_metadata = MetaData()
SFDC_SCHEMA_NAME = "sfdc"
manifest_file_path = "extract/sfdc/manifest.yaml"


class SfdcExtractor:
    """
    SaleForce Extractor
    """

    def __init__(self):
        self.name = "sfdc-extractor"
        self.tables = tables_from_manifest(
            manifest_file_path=manifest_file_path,
            metadata=sfdc_metadata,
            schema_name=SFDC_SCHEMA_NAME,
        )
        self.primary_keys = {
            table_name: [table.primary_key.name]
            for table_name, table in self.tables.items()
        }
        self.entities = [ent for ent in self.tables]

    def extract(self, entity) -> Iterator[pd.DataFrame]:
        pass



import logging
import subprocess
import asyncio

from .project import Project
from .plugin import Plugin
from .plugin.error import PluginMissingError
from .plugin.config_service import PluginConfigService
from .venv_service import VenvService


class PluginInvoker:
    """This class handles the invocation of a `Plugin` instance."""

    def __init__(
        self,
        project: Project,
        plugin: Plugin,
        run_dir=None,
        config_dir=None,
        venv_service: VenvService = None,
        config_service: PluginConfigService = None,
    ):
        self.project = project
        self.plugin = plugin
        self.venv_service = venv_service or VenvService(project)
        self.config_service = config_service or PluginConfigService(
            project, plugin, run_dir=run_dir, config_dir=config_dir
        )

        self._prepared = False

    @property
    def files(self):
        plugin_files = {**self.plugin.config_files, **self.plugin.output_files}

        return {
            _key: self.config_service.run_dir.joinpath(filename)
            for _key, filename in plugin_files.items()
        }

    def prepare(self):
        if not self._prepared:
            self.config_service.configure()
            self._prepared = True

    def exec_path(self):
        return self.venv_service.exec_path(self.plugin.name, namespace=self.plugin.type)

    def exec_args(self):
        plugin_args = self.plugin.exec_args(self.files)

        return [str(arg) for arg in (self.exec_path(), *plugin_args)]

    def invoke(self, *args, **Popen):
        try:
            self.prepare()
            with self.plugin.trigger_hooks("invoke", self, args):
                popen_args = [*self.exec_args(), *args]
                logging.debug(f"Invoking: {popen_args}")
                return subprocess.Popen(popen_args, **Popen)
        except Exception as err:
            logging.error(f"Failed to start plugin {self.plugin}.")
            raise PluginMissingError(self.plugin)

    async def invoke_async(self, *args, **Popen):
        try:
            self.prepare()
            with self.plugin.trigger_hooks("invoke", self, args):
                return await asyncio.create_subprocess_exec(
                    *self.exec_args(), *args, **Popen
                )

        except Exception as err:
            logging.error(f"Failed to start plugin {self.plugin}.")
            raise PluginMissingError(self.plugin)



import os

from sqlalchemy import create_engine
from sqlalchemy.dialects import postgresql
from sqlalchemy.schema import CreateSchema


class PostgresLoader:
    def __init__(self, **kwargs):
        """
        :param entity_name: str Name of the entity to load
        :param kwargs:
        """
        self.entity_name = kwargs["entity_name"]
        self.extractor = kwargs["extractor"]
        self.table = self.extractor.tables.get(self.entity_name)
        self.index_elements = self.extractor.primary_keys.get(self.entity_name)
        if "connection_string" in kwargs:
            self.engine = create_engine(kwargs["connection_string"])
        else:
            username = os.environ.get("PG_USERNAME")
            password = os.environ.get("PG_PASSWORD")
            host = os.environ.get("PG_ADDRESS")
            port = os.environ.get("PG_PORT")
            database = os.environ.get("PG_DATABASE")
            self.engine = create_engine(
                f"postgresql://{username}:{password}@{host}:{port}/{database}"
            )

    def schema_apply(self):
        # TODO: add loaded_at(timestamp) col to the schema
        if not self.engine.dialect.has_schema(self.engine, self.table.schema):
            print(f"Schema {self.table.schema} does not exist -> creating it ")
            self.engine.execute(CreateSchema(self.table.schema))
        # TODO: update table
        if not self.engine.dialect.has_table(
            self.engine, self.table.name, self.table.schema
        ):
            print(f"Table {self.table.name} does not exist -> creating it ")
            self.table.metadata.create_all(self.engine)

    def load(self, df):
        if not df.empty:
            dfs_to_load: list = df.to_dict(orient="records")
            insert_stmt = postgresql.insert(self.table).values(dfs_to_load)
            insert_stmt = insert_stmt.on_conflict_do_update(
                index_elements=self.index_elements,
                # only compatible with Postgres >= 9.5
                set_=insert_stmt.excluded._data,  # link to the conflicting data
            )
            print(f"Loading data to Postgres for {self.entity_name}")
            self.engine.execute(insert_stmt)
        else:
            print(f"DataFrame {df} is empty -> skipping it")



import re
import sys
import logging
from typing import Union

from requests.auth import HTTPBasicAuth
from functools import reduce
from pathlib import Path
from .db import DB


# from https://github.com/jonathanj/compose/blob/master/compose.py
def compose(*fs):
    """
    Create a function composition.
    :type *fs: ``iterable`` of 1-argument ``callable``s
    :param *fs: Iterable of 1-argument functions to compose, functions will be
        applied from last to first, in other words ``compose(f, g)(x) ==
        f(g(x))``.
    :return: I{callable} taking 1 argument.
    """
    return reduce(lambda f, g: lambda x: f(g(x)), fs, lambda x: x)


# from http://www.dolphmathews.com/2012/09/slugify-string-in-python.html
def slugify(s):
    """
    Simplifies ugly strings into something URL-friendly.
    >>> print slugify("[Some] _ Article's Title--")
    some-articles-title
    """

    # "[Some] _ Article's Title--"
    # "[some] _ article's title--"
    s = s.lower()

    # "[some] _ article's_title--"
    # "[some]___article's_title__"
    for c in [" ", "-", ".", "/"]:
        s = s.replace(c, "_")

    # "[some]___article's_title__"
    # "some___articles_title__"
    s = re.sub("\W", "", s)

    # "some___articles_title__"
    # "some   articles title  "
    s = s.replace("_", " ")

    # "some   articles title  "
    # "some articles title "
    s = re.sub("\s+", " ", s)

    # "some articles title "
    # "some articles title"
    s = s.strip()

    # "some articles title"
    # "some-articles-title"
    s = s.replace(" ", "-")

    return s


def setup_db(args=None):
    if args is None:
        DB.setup()
    else:
        DB.setup(**vars(args))


def setup_logging(log_level=logging.INFO):
    logging.basicConfig(
        stream=sys.stderr,
        format="[%(threadName)10s][%(levelname)s][%(asctime)s] %(message)s",
        level=log_level,
    )


def get_basic_auth(user, token):
    return HTTPBasicAuth(user, token)


def pop_all(keys, d: dict):
    return dict(map(lambda k: (k, d.pop(k)), keys))


def get_all(keys, d: dict, default=None):
    return dict(map(lambda k: (k, d.get(k, default)), keys))


def file_has_data(file: Union[Path, str]):
    file = Path(file)  # ensure it is a Path object
    return file.exists() and file.stat().st_size > 0


def noop(*_args, **_kwargs):
    pass



import pytest
import os
import json
from asynctest import CoroutineMock

from unittest import mock
from meltano.core.job import Job, State
from meltano.core.plugin import Plugin
from meltano.core.plugin_invoker import PluginInvoker
from meltano.core.plugin.singer import SingerTap, SingerTarget
from meltano.core.runner.singer import SingerRunner
from pathlib import Path


TEST_JOB_ID = "test_job"
TAP = SingerTap("tap-test")
TARGET = SingerTarget("target-test")


def create_plugin_files(config_dir: Path, plugin: Plugin):
    for file in plugin.config_files.values():
        Path(os.path.join(config_dir, file)).touch()

    return config_dir


class TestSingerRunner:
    @pytest.fixture()
    def subject(self, db_setup, mkdtemp, project):
        tap_config_dir = mkdtemp()
        target_config_dir = mkdtemp()

        create_plugin_files(tap_config_dir, TAP)
        create_plugin_files(target_config_dir, TARGET)

        Job(
            elt_uri=TEST_JOB_ID,
            state=State.SUCCESS,
            payload={"singer_state": {"bookmarks": []}},
        ).save()

        return SingerRunner(
            project,
            TEST_JOB_ID,
            tap_config_dir=tap_config_dir,
            target_config_dir=target_config_dir,
        )

    @pytest.fixture()
    def process_mock_factory(self):
        def _factory(name):
            process_mock = mock.Mock()
            process_mock.name = name
            process_mock.wait = CoroutineMock(return_value=0)
            return process_mock

        return _factory

    @pytest.fixture()
    def tap_process(self, process_mock_factory):
        tap = process_mock_factory(TAP)
        tap.stdout.readline = CoroutineMock(return_value="{}")
        return tap

    @pytest.fixture()
    def target_process(self, process_mock_factory):
        target = process_mock_factory(TARGET)
        return target

    def test_prepare_job(self, subject):
        tap_invoker = PluginInvoker(
            subject.project, TAP, config_dir=subject.tap_config_dir
        )
        target_invoker = PluginInvoker(
            subject.project, TARGET, config_dir=subject.target_config_dir
        )

        subject.prepare(tap_invoker, target_invoker)

        for f in TAP.config_files.values():
            assert subject.tap_config_dir.joinpath(f).exists()

        for f in TARGET.config_files.values():
            assert subject.target_config_dir.joinpath(f).exists()

    @pytest.mark.asyncio
    async def test_invoke(self, subject, tap_process, target_process):
        tap_invoker = PluginInvoker(
            subject.project, TAP, config_dir=subject.tap_config_dir
        )
        target_invoker = PluginInvoker(
            subject.project, TARGET, config_dir=subject.target_config_dir
        )

        # call prepare beforehand
        subject.prepare(tap_invoker, target_invoker)

        invoke_async = CoroutineMock(side_effect=(tap_process, target_process))

        with mock.patch.object(
            SingerRunner, "bookmark", new=CoroutineMock()
        ), mock.patch.object(
            PluginInvoker, "invoke_async", new=invoke_async
        ) as invoke_async:
            # async method
            await subject.invoke(tap_invoker, target_invoker)

            # correct bins are called
            assert invoke_async.awaited_with(tap_invoker)
            assert invoke_async.awaited_with(target_invoker)

            tap_process.wait.assert_awaited()
            target_process.wait.assert_awaited()

    @pytest.mark.asyncio
    async def test_bookmark(self, subject, tap_process, target_process):
        lines = (
            b'{"type": "STATE", "value": {"line": 1}}\n',
            b'{"type": "STATE", "value": {"line": 2}}\n',
            b'{"type": "STATE", "value": {"line": 3}}\n',
        )

        # testing with a real subprocess proved to be pretty
        # complicated.
        target_process.stdout = mock.MagicMock()
        target_process.stdout.at_eof.side_effect = (False, False, False, True)
        target_process.stdout.readline = CoroutineMock(side_effect=lines)

        with subject.job.run():
            await subject.bookmark(target_process.stdout)

        # assert the STATE's `value` was saved
        assert subject.job.payload["singer_state"] == {"line": 3}

        # test the restore
        tap_invoker = PluginInvoker(subject.project, TAP)
        subject.restore_bookmark(tap_invoker)
        state_json = json.dumps(subject.job.payload["singer_state"])
        assert tap_invoker.files["state"].exists()
        assert tap_invoker.files["state"].open().read() == state_json



import subprocess
import logging

from .venv_service import VenvService


class DbtService:
    def __init__(self, project, venv_service: VenvService = None):
        self.project = project
        self.venv_service = venv_service or VenvService(project)

    @property
    def exec_path(self):
        return self.venv_service.exec_path("dbt")

    def call(self, *args):
        logging.debug(f"Invoking: dbt {args}")
        exec_args = list(map(str, [self.exec_path, *args]))
        run = subprocess.run(exec_args)

        run.check_returncode()
        return run

    def compile(self):
        return self.call(
            "compile", "--profiles-dir", self.project.root, "--profile", "meltano"
        )

    def deps(self):
        return self.call("deps", "--profiles-dir", self.project.root)

    def run(self):
        return self.call(
            "run", "--profiles-dir", self.project.root, "--profile", "meltano"
        )



import pytest

from meltano.core.job import Job, State
from datetime import datetime


def sample_job(payload={}):
    return Job(elt_uri="elt://bizops/sample-elt", state=State.IDLE, payload=payload)


def test_save(session):
    subject = sample_job().save()

    assert subject.id > 0


def test_load(session):
    [session.add(sample_job({"key": i})) for i in range(0, 10)]

    subjects = session.query(Job).filter_by(elt_uri="elt://bizops/sample-elt")

    assert len(subjects.all()) == 10
    session.rollback()


def test_transit(session):
    subject = sample_job().save()

    transition = subject.transit(State.RUNNING)
    assert transition == (State.IDLE, State.RUNNING)
    subject.started_at = datetime.utcnow()

    transition = subject.transit(State.SUCCESS)
    assert transition == (State.RUNNING, State.SUCCESS)
    subject.ended_at = datetime.utcnow()


def test_run(session):
    subject = sample_job().save()

    # A successful run will mark the subject as SUCCESS and set the `ended_at`
    with subject.run():
        assert subject.state == State.RUNNING
        assert subject.ended_at is None

    assert subject.state == State.SUCCESS
    assert subject.ended_at is not None


def test_run_failed(session):
    # A failed run will mark the subject as FAILED an set the payload['error']
    subject = sample_job().save()
    exception = Exception("This is a test.")

    with pytest.raises(Exception) as e:
        with subject.run():
            raise exception

        # raise the same exception
        assert e is exception

    assert subject.state == State.FAIL
    assert subject.ended_at is not None
    assert subject.payload["error"] == "This is a test."


def test_this(session):
    assert session



import os
import yaml
import click
from urllib.parse import urlparse

from meltano.core.project_init_service import (
    ProjectInitService,
    ProjectInitServiceError,
)
from meltano.core.plugin_install_service import PluginInstallService
from . import cli

EXTRACTORS = "extractors"
LOADERS = "loaders"
ALL = "all"


@cli.command()
@click.argument("project_name")
def init(project_name):
    init_service = ProjectInitService(project_name)
    try:
        project = init_service.init()
        click.echo("Installing dbt...")
        init_service.install_dbt(project)
        init_service.echo_instructions()
    except ProjectInitServiceError as e:
        print(e)
        click.secho(f"Directory {project_name} already exists!", fg="red")
        raise click.Abort()



import click
import sys
from . import cli

from meltano.core.project import Project
from meltano.core.plugin_invoker import PluginInvoker
from meltano.core.config_service import ConfigService


@cli.command(context_settings=dict(ignore_unknown_options=True))
@click.argument("plugin_name")
@click.argument("plugin_args", nargs=-1, type=click.UNPROCESSED)
def invoke(plugin_name, plugin_args):
    project = Project.find()
    config_service = ConfigService(project)

    plugin = next(
        plugin for plugin in config_service.plugins() if plugin.name == plugin_name
    )

    service = PluginInvoker(project, plugin)
    handle = service.invoke(*plugin_args)

    try:
        exit_code = handle.wait()
        sys.exit(exit_code)
    except Exception as err:
        click.secho(f"An error occured: {err}.", fg="red")
        raise click.Abort() from err



import os
import yaml
from typing import Dict, List

from .plugin import Plugin, PluginType
from .plugin.singer import plugin_factory


class PluginNotFoundError(Exception):
    pass


class PluginDiscoveryInvalidError(Exception):
    invalid_message = "Invalid discovery file."
    pass


class PluginDiscoveryService:
    def __init__(self):
        self.discovery_file = os.path.join(os.path.dirname(__file__), "discovery.yml")
        self.load()

    def load(self):
        with open(self.discovery_file) as f:
            try:
                self.discovery_data = yaml.load(f)
            except Exception as e:
                raise PluginDiscoveryInvalidError()

    def plugins(self) -> List[Plugin]:
        """Parse the discovery file and returns it as `Plugin` instances."""
        # this will parse the discovery file and create an instance of the
        # corresponding `plugin_class` for all the plugins.
        return (
            plugin_factory(plugin_type, plugin_def)
            for plugin_type, plugin_defs in self.discovery_data.items()
            for plugin_def in plugin_defs
        )

    def find_plugin(self, plugin_type: PluginType, plugin_name: str):
        try:
            return next(
                plugin
                for plugin in self.plugins()
                if (plugin.type == plugin_type and plugin.name == plugin_name)
            )
        except StopIteration:
            raise PluginNotFoundError()

    def discover(self, plugin_type: PluginType):
        """Return a pretty printed list of available plugins."""
        enabled_plugin_types = (
            (PluginType.EXTRACTORS, PluginType.LOADERS)
            if PluginType.ALL
            else (plugin_type,)
        )
        return {
            plugin_type: self.list_discovery(plugin_type)
            for plugin_type in enabled_plugin_types
        }

    def list_discovery(self, discovery):
        return "\n".join(
            plugin.name for plugin in self.plugins() if plugin.type == discovery
        )



import uuid

from sqlalchemy.dialects.postgresql import UUID
from sqlalchemy.types import TypeDecorator, CHAR

from ..app import db


class GUID(TypeDecorator):
    """
    Platform-independent GUID type.
    Uses Postgresql's UUID type, otherwise uses
    CHAR(32), storing as stringified hex values.
    """

    impl = CHAR

    def load_dialect_impl(self, dialect):
        if dialect.name == "postgresql":
            return dialect.type_descriptor(UUID())
        else:
            return dialect.type_descriptor(CHAR(32))

    def process_bind_param(self, value, dialect):
        if value is None:
            return value
        elif dialect.name == "postgresql":
            return str(value)
        else:
            if not isinstance(value, uuid.UUID):
                return "%.32x" % uuid.UUID(value).int
            else:
                # hexstring
                return "%.32x" % value.int

    def process_result_value(self, value, dialect):
        if value is None:
            return value
        else:
            return uuid.UUID(value)


class Base(db.Model):
    __abstract__ = True

    id = db.Column(db.Integer, primary_key=True)
    date_created = db.Column(db.DateTime, default=db.func.current_timestamp())
    date_modified = db.Column(
        db.DateTime,
        default=db.func.current_timestamp(),
        onupdate=db.func.current_timestamp(),
    )



import pytest
import os
import contextlib
import sqlalchemy

from meltano.core.db import DB, SystemModel
from sqlalchemy import MetaData


@pytest.fixture(scope="session")
def db_setup(request):
    args = {
        "database": "pytest",
        "host": os.getenv("PG_ADDRESS"),
        "port": os.getenv("PG_PORT", 5432),
        "user": os.getenv("PG_USERNAME"),
        "password": os.getenv("PG_PASSWORD"),
    }
    DB.setup(**args)
    with contextlib.suppress(sqlalchemy.exc.ProgrammingError):
        DB.create_database("pytest")
    DB.default.ensure_schema_exists("meltano")

    # Create the base models
    SystemModel.metadata.create_all(DB.default.engine)

    truncate_tables(DB.default.engine, schema="meltano")


@pytest.fixture(scope="function")
def db(request, db_setup):
    def teardown():
        truncate_tables(DB.default, schema="meltano")

    request.addfinalizer(teardown)
    return DB.default


@pytest.fixture(scope="function")
def session(request, db):
    """Creates a new database session for a test."""
    session = db.create_session()

    request.addfinalizer(session.rollback)

    return session


def truncate_tables(db, schema):
    # delete all table data (but keep tables)
    # we do cleanup before test 'cause if previous test errored,
    # DB can contain dust
    con = db.engine.connect()
    trans = con.begin()
    con.execute("SET session_replication_role TO 'replica';")

    meta = MetaData(bind=db.engine, schema=schema)
    meta.reflect()
    for table in meta.sorted_tables:
        con.execute(table.delete())

    con.execute("SET session_replication_role TO 'origin';")
    trans.commit()



import yaml
import pytest
from click.testing import CliRunner
from unittest.mock import Mock, patch

from meltano.cli import cli
from meltano.core.project_add_service import ProjectAddService
from meltano.core.plugin_discovery_service import PluginDiscoveryService
from meltano.core.plugin_invoker import PluginInvoker
from meltano.core.plugin import PluginType


class TestCliInvoke:
    @pytest.fixture
    def subject(self, project_add_service, config_service):
        project_add_service.add(PluginType.EXTRACTORS, "tap-first")

        return CliRunner()

    @pytest.fixture
    def process_mock(self):
        process_mock = Mock()
        process_mock.wait.return_value = 0

        return process_mock

    def test_invoke(self, subject, process_mock):
        with patch.object(PluginInvoker, "invoke", return_value=process_mock) as invoke:
            basic = subject.invoke(cli, ["invoke", "tap-first"])
            assert invoke.called_once

    def test_invoke_args(self, subject, process_mock):
        with patch.object(PluginInvoker, "invoke", return_value=process_mock) as invoke:
            with_args = subject.invoke(cli, ["invoke", "tap-first", "--discover"])

            assert invoke.called_with(["--discover"])

    def test_invoke_exit_code(self, subject, process_mock):
        process_mock.wait.return_value = 2

        with patch.object(PluginInvoker, "invoke", return_value=process_mock) as invoke:
            basic = subject.invoke(cli, ["invoke", "tap-first"])
            assert basic.exit_code == 2



import datetime
import logging
import os
from logging.handlers import RotatingFileHandler

from flask import Flask, request
from flask import jsonify
from flask_cors import CORS
from flask_sqlalchemy import SQLAlchemy

from . import config
from .external_connector import ExternalConnector
from meltano.core.db import DB

app = Flask(__name__)

# unused afaik
meltano_model_path = config.meltano_model_path
meltano_transform_path = config.meltano_transform_path

app.config.from_object(config)

if os.environ["FLASK_ENV"] == "development":
    CORS(app)

# TODO: we need to setup proper dependency injection for
# this kind of hard-coupling. We are building multiple UI on
# the same backend. Almost all component need a ready DB
# connection.
DB.setup(
    host=app.config["MELTANO_POSTGRES_URL"],
    username=app.config["MELTANO_POSTGRES_USER"],
    password=app.config["MELTANO_POSTGRES_PASSWORD"],
    database=app.config["MELTANO_POSTGRES_DB"],
)
db = SQLAlchemy(app)

connector = ExternalConnector()

logger = logging.getLogger("melt_logger")
handler = RotatingFileHandler(app.config["LOG_PATH"], maxBytes=2000, backupCount=10)
logger.addHandler(handler)
now = str(datetime.datetime.utcnow().strftime("%b %d %Y %I:%M:%S:%f"))
logger.warning(f"Melt started at: {now}")


@app.before_request
def before_request():
    logger.info(f"[{request.remote_addr}] request: {now}")


@app.errorhandler(500)
def internal_error(exception):
    logger.info(f"[{request.remote_addr}] request: {now}, error: {exception}")
    return jsonify({"error": 1}), 500


@app.route("/")
def hello():
    return jsonify({"hello": 1})


from .controllers.projects import projectsBP
from .controllers.repos import reposBP
from .controllers.settings import settingsBP
from .controllers.sql import sqlBP

from .controllers.orchestrations import orchestrationsBP

app.register_blueprint(projectsBP)
app.register_blueprint(reposBP)
app.register_blueprint(settingsBP)
app.register_blueprint(sqlBP)
app.register_blueprint(orchestrationsBP)



import os
import json
import yaml
import logging

from .project import Project
from .plugin import PluginType, Plugin
from .plugin_discovery_service import PluginDiscoveryService
from .config_service import ConfigService


class PluginNotSupportedException(Exception):
    pass


class MissingPluginException(Exception):
    pass


class ProjectAddService:
    EXTRACTOR = "extractor"
    LOADER = "loader"

    def __init__(
        self,
        project: Project,
        discovery_service: PluginDiscoveryService = None,
        config_service: ConfigService = None,
    ):
        self.project = project
        self.discovery_service = discovery_service or PluginDiscoveryService()
        self.config_service = config_service or ConfigService(project)

    def add(self, plugin_type: PluginType, plugin_name: str):
        plugin = self.discovery_service.find_plugin(plugin_type, plugin_name)

        with self.project.meltano_update() as meltano_yml:
            meltano_yml[plugin_type] = meltano_yml.get(plugin_type, [])

        if plugin.pip_url:
            self.add_to_file(plugin)
        else:
            raise PluginNotSupportedException()

        return plugin

    def add_to_file(self, plugin: Plugin):
        if plugin in self.config_service.plugins():
            logging.warning(
                f"{plugin.name} is already present, use `meltano install` to install it."
            )
            return

        with self.project.meltano_update() as meltano_yml:
            meltano_yml[plugin.type].append(plugin.canonical())



import os
import pytest
import tempfile
import shutil
from pathlib import Path
from functools import partial


@pytest.fixture(scope="function")
def mkdtemp(request):
    def _mkdtemp(*args, **kwargs):
        path = Path(tempfile.mkdtemp())
        cleanup = partial(shutil.rmtree, path, ignore_errors=True)

        request.addfinalizer(cleanup)
        return path

    return _mkdtemp


@pytest.fixture(scope="session")
def test_dir(tmp_path_factory):
    test_dir = tmp_path_factory.mktemp("meltano_root")

    try:
        os.chdir(test_dir)
        yield test_dir
    finally:
        shutil.rmtree(test_dir)



def test_init(project):
    files = (
        project.root.joinpath(file).resolve()
        for file in (
            "meltano.yml",
            "README.md",
            ".gitignore",
            "dbt_project.yml",
            "packages.yml",
            "profiles.yml",
        )
    )

    dirs = (
        project.root.joinpath(dir)
        for dir in (
            ".meltano",
            "model",
            "extract",
            "load",
            "transform",
            "analyze",
            "notebook",
            "orchestrate",
        )
    )

    for file in files:
        assert file.is_file()

    for dir in dirs:
        assert dir.is_dir()



from .config_service import ConfigService
from .venv_service import VenvService
from .utils import noop
from .plugin import Plugin
from .project import Project
from .error import PluginInstallError, PluginInstallWarning


class PluginInstallService:
    def __init__(
        self,
        project: Project,
        venv_service: VenvService = None,
        config_service: ConfigService = None,
    ):
        self.project = project
        self.venv_service = venv_service or VenvService(project)
        self.config_service = config_service or ConfigService(project)

    def create_venv(self, plugin: Plugin):
        return self.venv_service.create(namespace=plugin.type, name=plugin.name)

    def install_all_plugins(self, status_cb=noop):
        errors = []
        installed = []

        for plugin in self.config_service.plugins():
            status = {"plugin": plugin, "status": "running"}
            status_cb(status)

            try:
                self.create_venv(plugin)
                self.install_plugin(plugin)

                status["status"] = "success"
                installed.append(status)
                status_cb(status)
            except PluginInstallError as err:
                status["status"] = "error"
                status["message"] = str(err)
                errors.append(status)
                status_cb(status)
            except PluginInstallWarning as warn:
                status["status"] = "warning"
                status["message"] = str(warn)
                errors.append(status)
                status_cb(status)

        return {"errors": errors, "installed": installed}

    def install_plugin(self, plugin: Plugin):
        try:
            with plugin.trigger_hooks("install", self.project):
                install_result = self.venv_service.install(
                    namespace=plugin.type, name=plugin.name, pip_url=plugin.pip_url
                )
                return install_result
        except Exception as err:
            raise PluginInstallError()



import json
from collections import OrderedDict
from datetime import date, datetime
from decimal import Decimal

from flask import Blueprint, jsonify, request

sqlBP = Blueprint("sql", __name__, url_prefix="/sql")

import sqlalchemy

from .utils import SqlHelper
from ..models.data import Model, Explore
from ..models.settings import Settings

connections = {}


def default(obj):
    if isinstance(obj, Decimal):
        return str(obj)
    elif isinstance(obj, (datetime, date)):
        return obj.isoformat()
    raise TypeError(f"Object of type {type(obj).__name__} is not JSON serialize-able")


def update_connections():
    current_connections = Settings.query.first().settings["connections"]
    for connection in current_connections:
        connection_name = connection["name"]
        if connection_name not in connections:
            this_connection = {}
            if connection["dialect"] == "postgresql":
                psql_params = ["username", "password", "host", "port", "database"]
                user, pw, host, port, db = [connection[param] for param in psql_params]
                connection_url = f"postgresql+psycopg2://{user}:{pw}@{host}:{port}/{db}"
                this_connection["connection_url"] = connection_url
                this_connection["engine"] = sqlalchemy.create_engine(
                    this_connection["connection_url"]
                )
            connections[connection_name] = this_connection


@sqlBP.route("/", methods=["GET"])
def index():
    return jsonify({"result": True})


@sqlBP.route("/get/<model_name>/<explore_name>", methods=["POST"])
def get_sql(model_name, explore_name):
    update_connections()
    sqlHelper = SqlHelper()
    model = Model.query.filter(Model.name == model_name).first()
    explore = Explore.query.filter(Explore.name == explore_name).first()
    incoming_json = request.get_json()
    to_run = incoming_json["run"]
    sql_dict = sqlHelper.get_sql(explore, incoming_json)
    outgoing_sql = sql_dict["sql"]
    incoming_order = incoming_json["order"]
    incoming_order_desc = incoming_json["desc"]
    measures = sql_dict["measures"]
    dimensions = sql_dict["dimensions"]
    column_headers = sql_dict["column_headers"]

    if to_run:
        db_to_connect = model.settings["connection"]
        if not db_to_connect in connections:
            return (
                jsonify(
                    {
                        "error": True,
                        "code": f"Missing connection details to {db_to_connect}. Create a connection to {db_to_connect} in the settings.",
                    }
                ),
                422,
            )
        engine = connections[model.settings["connection"]]["engine"]

        try:
            results = engine.execute(outgoing_sql)
        except sqlalchemy.exc.DBAPIError as e:
            print(e)
            return (
                jsonify(
                    {
                        "error": True,
                        "code": e.code,
                        "orig": e.orig.diag.message_primary,
                        "statement": e.statement,
                    }
                ),
                422,
            )

        results = [OrderedDict(row) for row in results]
        base_dict = {"sql": outgoing_sql, "results": results, "error": False}
        if not len(results):
            base_dict["empty"] = True
        else:
            base_dict["empty"] = False
            base_dict["column_headers"] = column_headers
            base_dict["keys"] = list(results[0].keys())
            base_dict["measures"] = sqlHelper.get_names(measures)

        return json.dumps(base_dict, default=default)
    else:
        return json.dumps({"sql": outgoing_sql}, default=default)


@sqlBP.route("/distinct/<model_name>/<explore_name>", methods=["POST"])
def get_distinct_field_name(model_name, explore_name):
    update_connections()
    incoming_json = request.get_json()
    field_name = incoming_json["field"].replace("${TABLE}", explore_name)
    model = Model.query.filter(Model.name == model_name).first()
    explore = Explore.query.filter(Explore.name == explore_name).first()
    base_table = explore.view.settings["sql_table_name"]
    base_sql = f"SELECT DISTINCT {field_name} FROM {base_table} AS {explore_name} ORDER BY {field_name}"
    engine = connections[model.settings["connection"]]["engine"]
    results = engine.execute(base_sql)
    results = [dict(row) for row in results]
    return json.dumps(
        {"sql": base_sql, "results": results, "keys": list(results[0].keys())},
        default=default,
    )



import os
import logging
import pandas as pd

from sqlalchemy import (
    create_engine,
    inspect,
    MetaData,
    Table,
    Column,
    String,
    Integer,
    Float,
)
from sqlalchemy.schema import CreateSchema

from snowflake.sqlalchemy import URL
from sqlalchemy.sql import select

if __name__ == "__main__":
    # Set proper Logging Level
    Log = logging.getLogger()
    level = logging.getLevelName("INFO")
    Log.setLevel(level)

    # Create engine for Snowflake
    engine = create_engine(
        URL(
            user=os.environ.get("SNOWFLAKE_USERNAME"),
            password=os.environ.get("SNOWFLAKE_PASSWORD"),
            account=os.environ.get("SNOWFLAKE_ACCOUNT_NAME"),
            database=os.environ.get("SNOWFLAKE_DB_NAME"),
            warehouse=os.environ.get("SNOWFLAKE_WAREHOUSE"),
        )
    )

    db_name = os.environ.get("SNOWFLAKE_DB_NAME")

    try:
        logging.info(
            "Testing that the connection to Snowflake has been properly set up."
        )

        logging.info("Connecting to Snowflake")
        connection = engine.connect()

        logging.info("Executing simple query")
        results = connection.execute("select current_version()").fetchone()
        logging.info(f"Current Snowflake version: {results[0]}")

        demo_metadata = MetaData()
        test_table = Table(
            "test_table",
            demo_metadata,
            Column("id", Integer, primary_key=True),
            Column("id2", String, primary_key=True),
            Column("name", String),
            Column("value", Float),
            schema="meltano_loader_tests",
        )

        logging.info("Checking if test schema and test table exist")

        inspector = inspect(engine)

        all_schema_names = inspector.get_schema_names()
        if not (test_table.schema in all_schema_names):
            logging.info(f"Schema {test_table.schema} does not exist -> creating it ")
            engine.execute(CreateSchema(test_table.schema))
        else:
            logging.info(f"Schema {test_table.schema} found!")

        all_table_names = inspector.get_table_names(test_table.schema)
        if not (test_table.name in all_table_names):
            logging.info(f"Table {test_table.name} does not exist -> creating it ")
            test_table.metadata.create_all(engine)
        else:
            logging.info(f"Table {test_table.name} found!")

        logging.info(f"Creating temporary table tmp_{test_table.name}")
        columns = [c.copy() for c in test_table.columns]
        tmp_table = Table(
            f"tmp_{test_table.name}",
            test_table.metadata,
            *columns,
            schema=test_table.schema,
        )

        tmp_table.drop(engine, checkfirst=True)
        tmp_table.create(engine)

        logging.info(f"Loading data to table {tmp_table.name}")

        test_data = [
            {"id": 1, "id2": "a", "name": "test1a", "value": 1.21},
            {"id": 1, "id2": "b", "name": "test1b", "value": 1.26},
            {"id": 2, "id2": "a", "name": "test2a", "value": 2.96},
            {"id": 3, "id2": "a", "name": "test3a", "value": 3.04},
            {"id": 3, "id2": "b", "name": "test3b", "value": 3.09},
            {"id": 4, "id2": "a", "name": "test4a", "value": 4.57},
        ]

        df = pd.DataFrame(data=test_data)
        dfs_to_load: list = df.to_dict(orient="records")

        connection.execute(tmp_table.insert(), dfs_to_load)

        logging.info(f"Merging {tmp_table.name} into {test_table.name}")

        merge_target = f"{test_table.schema}.{test_table.name}"
        merge_source = f"{test_table.schema}.{tmp_table.name}"

        joins = []
        for primary_key in test_table.primary_key:
            pk = primary_key.name
            joins.append(f"{merge_target}.{pk} = {merge_source}.{pk}")
        join_expr = " AND ".join(joins)

        attributes_target = []
        attributes_source = []
        update_sub_clauses = []

        for column in test_table.columns:
            attr = column.name
            attributes_target.append(attr)
            attributes_source.append(f"{merge_source}.{attr}")

            if attr not in test_table.primary_key:
                update_sub_clauses.append(f"{attr} = {merge_source}.{attr}")

        update_clause = ", ".join(update_sub_clauses)
        matched_clause = f"WHEN MATCHED THEN UPDATE SET {update_clause}"

        source_attributes = ", ".join(attributes_target)
        target_attributes = ", ".join(attributes_source)

        insert_clause = f"({source_attributes}) VALUES ({target_attributes})"
        not_matched_clause = f"WHEN NOT MATCHED THEN INSERT {insert_clause}"

        merge_stmt = f"MERGE INTO {merge_target} USING {merge_source} ON {join_expr} {matched_clause} {not_matched_clause}"
        connection.execute(merge_stmt)

        logging.info(f"Fetching back the data from table {tmp_table.name}")
        results = connection.execute(select([tmp_table]).order_by("id", "id2"))
        for row in results:
            logging.info(f"{row['id']} - {row['id2']} - {row['name']} - {row['value']}")
        results.close()

        logging.info(f"Fetching back the data from table {test_table.name}")
        results = connection.execute(select([test_table]).order_by("id", "id2"))
        for row in results:
            logging.info(f"{row['id']} - {row['id2']} - {row['name']} - {row['value']}")
        results.close()

        logging.info("Droping Schema and Tables")
        tmp_table.drop(engine)
        test_table.drop(engine)
        connection.execute(f"DROP SCHEMA {db_name}.{test_table.schema} CASCADE")

        # Facilitate Testing when the tables are not dropped
        # grant_stmt = f'GRANT USAGE ON SCHEMA {db_name}.{test_table.schema} TO ROLE LOADER;'
        # connection.execute(grant_stmt)
        # grant_stmt = f'GRANT SELECT ON ALL TABLES IN SCHEMA {db_name}.{test_table.schema} TO ROLE LOADER;'
        # connection.execute(grant_stmt)

        logging.info("Success!")
    finally:
        connection.close()
        engine.dispose()



import os
import yaml
from typing import List

from .project import Project
from .plugin import Plugin, PluginType
from .plugin.singer import plugin_factory
from .plugin.error import PluginMissingError


class ConfigService:
    def __init__(self, project: Project):
        self.project = project

    def make_meltano_secret_dir(self):
        os.makedirs(self.project.meltano_dir(), exist_ok=True)

    def get_plugin(self, plugin_type: PluginType, plugin_name: str):
        try:
            return next(
                plugin
                for plugin in self.plugins()
                if plugin.type == plugin_type and plugin.name == plugin_name
            )
        except StopIteration:
            raise PluginMissingError(plugin_name)

    def get_extractors(self):
        return filter(lambda p: p.type == PluginType.EXTRACTORS, self.plugins())

    def get_loaders(self):
        return filter(lambda p: p.type == PluginType.LOADERS, self.plugins())

    def get_database(self, database_name):
        return yaml.load(
            open(self.project.meltano_dir(f".database_{database_name}.yml"))
        )

    def plugins(self) -> List[Plugin]:
        """Parse the discovery file and returns it as `Plugin` instances."""
        # this will parse the discovery file and create an instance of the
        # corresponding `plugin_class` for all the plugins.
        return (
            plugin_factory(plugin_type, plugin_def)
            for plugin_type, plugin_defs in self.project.meltano.items()
            for plugin_def in plugin_defs
        )



from meltano.core.db import DB, SystemModel


class Runner:
    def before_run(self, *args, **kwargs):
        DB.default.ensure_schema_exists("meltano")
        SystemModel.metadata.create_all(DB.default.engine)

    def run(self, *args, **kwargs):
        pass

    def perform(self, *args, **kwargs):
        self.before_run(*args, **kwargs)
        self.run(*args, **kwargs)



import json

from meltano.core.plugin.config_service import PluginConfigService


class TestPluginConfigService:
    subject_cls = PluginConfigService

    def test_envsubst(self, tmp_path):
        src = tmp_path.joinpath("src")
        dst = tmp_path.joinpath("dst")

        with src.open("w") as o:
            o.write(
                "{"
                '  "var": "$VAR",'
                '  "foo": "${FOO}",'
                '  "missing": "$MISSING",'
                '  "multiple": "$A ${B} $C"'
                "}"
            )

        env = {
            "VAR": "hello world!",
            "FOO": 42,
            "A": "rock",
            "B": "paper",
            "C": "scissors",
        }

        self.subject_cls.envsubst(src, dst, env=env)

        parsed_dst = json.load(dst.open())
        assert parsed_dst["var"] == env["VAR"]
        assert parsed_dst["foo"] == str(env["FOO"])
        assert parsed_dst["missing"] == ""
        assert parsed_dst["multiple"] == "rock paper scissors"



import os
import functools
import click

from meltano.core.utils import pop_all
from meltano.core.db import DB


def db_options(func):
    @click.option(
        "-H",
        "--host",
        envvar="PG_ADDRESS",
        default="localhost",
        help="Database address.",
    )
    @click.option("-p", "--port", type=int, envvar="PG_PORT", default=5432)
    @click.option(
        "-d",
        "-db",
        "database",
        envvar="PG_DATABASE",
        help="Database to import the data to.",
    )
    @click.option(
        "-u",
        "--username",
        envvar="PG_USERNAME",
        default=lambda: os.getenv("USER", ""),
        help="Specifies the user to connect to the database with.",
    )
    @click.password_option(envvar="PG_PASSWORD")
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        config = pop_all(("host", "port", "database", "username", "password"), kwargs)
        DB.setup(**config)
        return func(*args, **kwargs)

    return wrapper



import click
import logging
import warnings

# disable the psycopg2 warning
# this needs to run before `psycopg2` is imported
warnings.filterwarnings("ignore", category=UserWarning, module="psycopg2")

from meltano.core.utils import setup_logging


LEVELS = {
    "debug": logging.DEBUG,
    "info": logging.INFO,
    "warning": logging.WARNING,
    "error": logging.ERROR,
    "critical": logging.CRITICAL,
}


@click.group(invoke_without_command=True)
@click.option("--log-level", type=click.Choice(LEVELS.keys()), default="info")
@click.pass_context
def cli(ctx, log_level):
    setup_logging(log_level=LEVELS[log_level])



import uuid

from .base import Base
from ..app import db


class BaseLook(Base):
    __abstract__ = True
    name = db.Column(db.String(128), nullable=False)
    settings = db.Column(db.JSON(), nullable=True)

    def __init__(self, name, settings):
        self.name = name
        self.settings = settings


class Model(BaseLook):
    __tablename__ = "model"

    def __init__(self, name, settings):
        super().__init__(name, settings)

    def __repr__(self):
        return "<Model %i>" % (self.id)


class Explore(BaseLook):
    __tablename__ = "explore"

    model_id = db.Column(db.Integer, db.ForeignKey("model.id"), nullable=False)
    model = db.relationship("Model", backref=db.backref("explores", lazy=True))
    view_id = db.Column(db.Integer, db.ForeignKey("view.id"), nullable=False)
    view = db.relationship("View", backref=db.backref("explores", lazy=True))
    joins = db.relationship("Join", backref="explore", cascade="all,delete", lazy=True)

    def __init__(self, name, settings):
        super().__init__(name, settings)

    def __repr__(self):
        return "<Explore %i>" % (self.id)

    def serializable(self, include_dimensions_and_measures=False):
        this_explore = {}
        this_explore["settings"] = self.settings
        this_explore["name"] = self.name
        this_explore_model = {}
        this_explore_model["settings"] = self.model.settings
        this_explore_model["name"] = self.model.name
        this_explore["model"] = this_explore_model
        this_explore["joins"] = []
        this_explore["unique_name"] = f"explore_{self.name}_{uuid.uuid4()}"
        this_view = {}
        this_view["name"] = self.view.name
        this_view["settings"] = self.view.settings
        this_view["unique_name"] = f"view_{self.view.name}_{uuid.uuid4()}"
        this_view["collapsed"] = True
        if include_dimensions_and_measures:
            this_view["dimensions"] = []
            for dimension in self.view.dimensions:
                this_dimension = {}
                this_dimension["name"] = dimension.name
                this_dimension["settings"] = dimension.settings
                this_dimension["label"] = dimension.settings.get(
                    "label", " ".join(dimension.name.split("_")).title()
                )
                this_dimension["view_label"] = dimension.settings.get(
                    "view_label", " ".join(dimension.name.split("_")).title()
                )
                this_dimension[
                    "unique_name"
                ] = f"dimension_{dimension.name}_{uuid.uuid4()}"
                this_dimension["selected"] = False
                this_view["dimensions"].append(this_dimension)

            this_view["dimension_groups"] = []
            for dimension_group in self.view.dimension_groups:
                this_dimension_group = {}
                this_dimension_group["name"] = dimension_group.name
                this_dimension_group["settings"] = dimension_group.settings
                this_dimension_group["label"] = dimension_group.settings.get(
                    "label", " ".join(dimension_group.name.split("_")).title()
                )
                this_dimension_group[
                    "unique_name"
                ] = f"dimension_group_{dimension_group.name}_{uuid.uuid4()}"
                this_dimension_group["selected"] = False
                this_dimension_group["timeframes"] = [
                    {
                        "label": tf.title().replace("*", ""),
                        "name": tf,
                        "selected": False,
                    }
                    for tf in dimension_group.settings["timeframes"]
                ]
                this_view["dimension_groups"].append(this_dimension_group)

            this_view["measures"] = []
            for measure in self.view.measures:
                this_measure = {}
                this_measure["name"] = measure.name
                this_measure["label"] = measure.settings.get(
                    "label", " ".join(measure.name.split("_")).title()
                )
                this_measure["settings"] = measure.settings
                this_measure["unique_name"] = f"measure_{measure.name}_{uuid.uuid4()}"
                this_measure["selected"] = False
                this_view["measures"].append(this_measure)
            this_explore["view"] = this_view
        for join in self.joins:
            this_join = {}
            this_join["name"] = join.name
            this_join["collapsed"] = True
            this_join["dimensions"] = []
            this_join["settings"] = join.settings
            this_join["settings"]["view_label"] = this_join["settings"].get(
                "view_label", " ".join(this_join["name"].split("_")).title()
            )
            this_explore["joins"].append(this_join)
        this_explore["joins"] = sorted(
            this_explore["joins"], key=lambda x: x["settings"]["view_label"]
        )
        return this_explore


class Join(BaseLook):
    __tablename__ = "join"

    explore_id = db.Column(db.Integer, db.ForeignKey("explore.id"), nullable=False)

    def __init__(self, name, settings):
        super().__init__(name, settings)


class View(BaseLook):
    __tablename__ = "view"

    def __init__(self, name, settings):
        super().__init__(name, settings)

    def __repr__(self):
        return "<View %i>" % (self.id)

    def serializable(self, include_dimensions_and_measures=False):
        this_view = {}
        this_view["name"] = self.name
        this_view["settings"] = self.settings
        this_view["dimensions"] = []
        for dimension in self.dimensions:
            this_dimension = {}
            this_dimension["name"] = dimension.name
            this_dimension["settings"] = dimension.settings
            this_dimension["label"] = dimension.settings.get(
                "label", " ".join(dimension.name.split("_")).title()
            )
            this_dimension["unique_name"] = f"dimension_{dimension.name}_{uuid.uuid4()}"
            this_dimension["selected"] = False
            this_view["dimensions"].append(this_dimension)
        this_view["measures"] = []
        for measure in self.measures:
            this_measure = {}
            this_measure["name"] = measure.name
            this_measure["label"] = measure.settings.get(
                "label", " ".join(measure.name.split("_")).title()
            )
            this_measure["settings"] = measure.settings
            this_measure["unique_name"] = f"measure_{measure.name}_{uuid.uuid4()}"
            this_measure["selected"] = False
            this_view["measures"].append(this_measure)
        return this_view


class Dimension(BaseLook):
    __tablename__ = "dimension"

    view_id = db.Column(db.Integer, db.ForeignKey("view.id"), nullable=False)
    view = db.relationship("View", backref=db.backref("dimensions", lazy=True))

    @property
    def table_column_name(self):
        if "sql" in self.settings:
            return self.settings["sql"].replace("${TABLE}", self.view.name).rstrip()
        return f"{self.view.name}.{self.name}".rstrip()

    @property
    def label(self):
        return self.settings.get("label", self.name.replace("_", " ").title())

    def __init__(self, name, settings):
        super().__init__(name, settings)

    def __repr__(self):
        return "<Dimension %i>" % (self.id)


class DimensionGroup(BaseLook):
    __tablename__ = "dimension_group"

    view_id = db.Column(db.Integer, db.ForeignKey("view.id"), nullable=False)
    view = db.relationship("View", backref=db.backref("dimension_groups", lazy=True))

    def __init__(self, name, settings):
        super().__init__(name, settings)

    @property
    def table_column_name(self):
        if "sql" in self.settings:
            return self.settings["sql"].replace("${TABLE}", self.view.name).rstrip()
        return f"{self.view.name}.{self.name}".rstrip()

    def __repr__(self):
        return "<DimensionGroup %i>" % (self.id)


class Measure(BaseLook):
    __tablename__ = "measure"

    view_id = db.Column(db.Integer, db.ForeignKey("view.id"), nullable=False)
    view = db.relationship("View", backref=db.backref("measures", lazy=True))

    @property
    def label(self):
        return self.settings.get("label", self.name.replace("_", " ").title())

    def __init__(self, name, settings):
        super().__init__(name, settings)

    def __repr__(self):
        return "<Measure %i>" % (self.id)



from flask import Blueprint, jsonify, request

settingsBP = Blueprint("settings", __name__, url_prefix="/settings")

from ..app import db

from ..models.settings import Settings


@settingsBP.route("/", methods=["GET"])
def index():
    settings = Settings.query.first()
    if not settings:
        settings = Settings()
    return jsonify(settings.serializable())


@settingsBP.route("/new", methods=["POST"])
def new():
    settings = request.get_json()
    current_settings = Settings.query.first()
    if not current_settings:
        current_settings = Settings()
    current_settings.settings = settings
    db.session.add(current_settings)
    db.session.commit()
    return jsonify(settings)


@settingsBP.route("/connections/<name>/test")
def test(name):
    current_settings = Settings.query.first().settings
    connections = current_settings["connections"]
    try:
        found_connection = next(
            connection for connection in connections if connection["name"] == name
        )
    # this is a really broad exception catch, this will swallow sneaky errors
    except Exception as e:
        found_connection = {}

    return jsonify(found_connection)



import os
import json
import click
from urllib.parse import urlparse

from . import cli
from meltano.core.plugin import PluginType
from meltano.core.plugin_discovery_service import (
    PluginDiscoveryService,
    PluginDiscoveryInvalidError,
)


@cli.command()
@click.argument(
    "plugin_type",
    type=click.Choice([PluginType.EXTRACTORS, PluginType.LOADERS, PluginType.ALL]),
)
def discover(plugin_type):
    discover_service = PluginDiscoveryService()
    try:
        discovery_dict = discover_service.discover(plugin_type)
        for key, value in discovery_dict.items():
            click.secho(key, fg="green")
            click.echo(value)
    except PluginDiscoveryInvalidError:
        click.secho(PluginDiscoveryInvalidError.invalid_message, fg="red")
        raise click.Abort()



import logging
import re

import sqlalchemy
from flask import jsonify
from pypika import Query, Table

from .aggregate import Aggregate
from .date import Date
from .join import Join as JoinQuery
from .substitution import Substitution
from ..app import app, db
from ..models.data import View, DimensionGroup, Join
from ..models.projects import Project
from ..models.settings import Settings


class SqlHelper:
    def parse_sql(self, input):
        placeholders = self.placeholder_match(input)

    def placeholder_match(self, input):
        outer_pattern = r"(\$\{[\w\.]*\})"
        inner_pattern = r"\$\{([\w\.]*)\}"
        outer_results = re.findall(outer_pattern, input)
        inner_results = re.findall(inner_pattern, input)
        return (outer_results, inner_results)

    def get_names(self, things):
        return [thing.name for thing in things]

    def get_sql(self, explore, incoming_json):
        view_name = incoming_json["view"]
        view = View.query.filter(View.name == view_name).first()
        base_table = view.settings["sql_table_name"]
        (schema, table) = base_table.split(".")
        incoming_dimensions = incoming_json["dimensions"]
        incoming_dimension_groups = incoming_json["dimension_groups"]
        incoming_measures = incoming_json["measures"]
        incoming_filters = incoming_json["filters"]
        incoming_joins = incoming_json["joins"]
        incoming_limit = incoming_json.get("limit", 50)
        # get all timeframes
        timeframes = [t["timeframes"] for t in incoming_dimension_groups]
        # flatten list of timeframes
        timeframes = [y for x in timeframes for y in x]
        dimensions = list(
            filter(lambda x: x.name in incoming_dimensions, view.dimensions)
        )
        measures = list(filter(lambda x: x.name in incoming_measures, view.measures))
        dimensions_raw = dimensions
        measures_raw = measures
        table = self.table(base_table, explore.name)
        # joins = self.joins(incoming_joins, table)
        dimension_groups = self.dimension_groups(
            view_name, incoming_dimension_groups, table
        )
        dimensions = self.dimensions(dimensions, table)
        dimensions = dimensions + dimension_groups
        measures = self.measures(measures, table)
        column_headers = self.column_headers(dimensions_raw, measures_raw)
        return {
            "dimensions": dimensions_raw,
            "measures": measures_raw,
            "column_headers": column_headers,
            "sql": self.get_query(
                from_=table,
                dimensions=dimensions,
                measures=measures,
                limit=incoming_limit,
            ),
        }

    def column_headers(self, dimensions, measures):
        return [d.label for d in dimensions + measures]

    def table(self, name, alias):
        (schema, name) = name.split(".")
        return Table(name, schema=schema, alias=alias)

    def dimensions(self, dimensions, table):
        return [self.field_from_dimension(d, table) for d in dimensions]

    def field_from_dimension(self, d, table):
        sql = d.settings["sql"]
        substitution = Substitution(sql, table, dimension=d)
        return substitution.sql

    def measures(self, measures, table):
        return [self.field_from_measure(m, table) for m in measures]

    def field_from_measure(self, m, table):
        aggregate = Aggregate(m, table)
        return aggregate.sql

    def joins(self, joins, table):
        if len(joins):
            for join in joins:
                queried_join = Join.query.filter(Join.name == join["name"]).first()
                join = JoinQuery(queried_join, view, table)

    def dimension_groups(self, view_name, dimension_groups, table):
        fields = []
        for dimension_group in dimension_groups:
            dimension_group_queried = (
                DimensionGroup.query.join(View, DimensionGroup.view_id == View.id)
                .filter(View.name == view_name)
                .filter(DimensionGroup.name == dimension_group["name"])
                .first()
            )
            (_table, name) = dimension_group_queried.table_column_name.split(".")
            for timeframe in dimension_group["timeframes"]:
                d = Date(timeframe, table, name)
                fields.append(d.sql)
        return fields

    def get_query(self, from_, dimensions, measures, limit):
        select = dimensions + measures
        q = Query.from_(from_).select(*select).groupby(*dimensions).limit(limit)
        return f"{str(q)};"

    @app.route("/drop_it_like_its_hot")
    def reset_db(self):
        try:
            Settings.__table__.drop(db.engine)
            Project.__table__.drop(db.engine)
            db.drop_all()
        except sqlalchemy.exc.OperationalError as err:
            logging.error("Failed drop database.")
        db.create_all()
        settings = Settings()
        db.session.add(settings)
        db.session.commit()
        return jsonify({"dropped_it": "like_its_hot"})



import yaml
from enum import Enum


class YAMLEnum(str, Enum):
    def __str__(self):
        return self.value

    @staticmethod
    def yaml_representer(dumper, obj):
        return dumper.represent_scalar("tag:yaml.org,2002:str", str(obj))


yaml.add_multi_representer(YAMLEnum, YAMLEnum.yaml_representer)


class PluginType(YAMLEnum):
    EXTRACTORS = "extractors"
    LOADERS = "loaders"
    ALL = "all"

    def __str__(self):
        return self.value


class Plugin:
    def __init__(
        self, plugin_type: PluginType, name: str, pip_url=None, config=None, **extras
    ):
        self.name = name
        self.type = plugin_type
        self.pip_url = pip_url
        self.config = config
        self._extras = extras or {}

    def canonical(self):
        return {
            "name": self.name,
            "pip_url": self.pip_url,
            "config": self.config,
            **self._extras,
        }

    @property
    def config_files(self):
        """Return a list of stubbed files created for this plugin."""
        return []

    @property
    def output_files(self):
        return []

    def __eq__(self, other):
        return self.name == other.name and self.type == other.type



import logging
import click
import psycopg2
import psycopg2.sql

from meltano.core.db import db_open
from . import cli
from .params import db_options


@cli.group()
@db_options
def schema():
    pass


@schema.command()
@click.argument("schema")
@click.argument("roles", nargs=-1, required=True)
def create(schema, roles):
    with db_open() as db:
        db.ensure_schema_exists(schema, grant_roles=roles)



import re
from collections import namedtuple
from enum import Enum

from pypika import Field, Case


class SubstitutionType(Enum):
    unknown = "UNKNOWN"
    table = "TABLE"
    dimension = "DIMENSION"
    view_dimension = "VIEW_DIMENSION"
    view_sql_table_name = "VIEW_SQL_TABLE_NAME"


class Substitution:
    def __init__(self, _input, table, dimension=None):
        self.input = _input
        self.alias = None
        self.sql = None
        self.table = table

        if not dimension:
            self.type = "string"
        else:
            self.type = dimension.settings["type"]

        self.substitution_type = SubstitutionType.unknown
        self.get_substitution_type()
        self.placeholders = Substitution.placeholder_match(self.input)
        self.set_sql()

    def get_substitution_type(self):
        # trying guess the substitution_type in a cheap way
        if "." in self.input and "${TABLE}" not in self.input:
            if "SQL_TABLE_NAME" in self.input:
                self.substitution_type = SubstitutionType.view_sql_table_name
            else:
                self.substitution_type = SubstitutionType.view_dimension
        elif "${TABLE}" in self.input:
            self.substitution_type = SubstitutionType.table
        elif " " not in self.input:
            self.substitution_type = SubstitutionType.dimension
        else:
            self.substitution_type = SubstitutionType.unknown

    @staticmethod
    def placeholder_match(input_):
        outer_pattern = r"(\$\{[\w\.]*\})"
        inner_pattern = r"\$\{([\w\.]*)\}"
        outer_results = re.findall(outer_pattern, input_)
        inner_results = re.findall(inner_pattern, input_)
        Results = namedtuple("Results", "inner outer")
        return Results(inner=inner_results, outer=outer_results)

    def set_sql(self):
        if self.substitution_type is SubstitutionType.table:
            self.set_sql_table_type()
        else:
            raise Exception(
                f"Substitution Type {self.substitution_type.value} not implemented yet"
            )

    def set_sql_table_type(self):
        self.sql = self.input.replace(
            self.placeholders.outer[0], self.table._table_name
        )
        (table, field) = self.sql.split(".")
        self.alias = self.sql
        if self.type == "yesno":
            field = Field(field, table=self.table)
            self.sql = Case(alias=self.sql).when(field, "yes").else_("no")
        else:
            self.sql = Field(field, table=self.table, alias=self.sql)



from ..app import db
from ..models.base import Base


class Project(Base):
    __tablename__ = "project"
    name = db.Column(db.String(128), nullable=False)
    git_url = db.Column(db.String(), nullable=False)
    validated = db.Column(db.Boolean(), default=False)

    def __init__(self, name, git_url):
        self.name = name
        self.git_url = git_url

    def __repr__(self):
        return f"<Project {self.name}>"



#!/usr/bin/env python
from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

requires = [
    'aiohttp',
    'backoff',
    'click',
    'dbt',
    'gitpython',
    'markdown',
    'pandas',
    'psycopg2',
    'pypika',
    'python-dotenv',
    'pyyaml',
    'snowflake-connector-python',
    'snowflake-sqlalchemy',
    'sqlalchemy',
]

api_requires = [
    'flask',
    'flask-cors',
    'flask-sqlalchemy',
]

dev_requires = [
    'pytest',
    'pytest-asyncio',
    'asynctest',
    'black',
    'bumpversion',
]

setup(
    name="meltano",
    version="0.1.3",
    author='Meltano Team & Contributors',
    author_email="meltano@gitlab.com",
    description="Meltano",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://gitlab.com/meltano/meltano",
    package_dir={'': 'src'},
    packages=find_packages(where="src"),
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    setup_requires=['pytest-runner'],
    tests_require=dev_requires,
    # run `make requirements.txt` after editing
    install_requires=requires,
    extras_require={
        'api': api_requires,
        'dev': dev_requires,
        'all': [
            *api_requires,
            *dev_requires,
        ],
    },
    entry_points={
        'console_scripts': [
            "meltano = meltano.cli:main"
        ]
    }
)



from flask import Blueprint, jsonify, request

projectsBP = Blueprint("projectsBP", __name__, url_prefix="/projects")

from ..app import db
from ..models.projects import Project
from ..models.settings import Settings


@projectsBP.route("/", methods=["GET"])
def index():
    p = Project.query.first()
    return (
        jsonify({"name": p.name, "git_url": p.git_url})
        if p
        else jsonify({"name": "", "git_url": ""})
    )


@projectsBP.route("/new", methods=["POST"])
def add():
    incoming = request.get_json()
    name = incoming.get("name")
    git_url = incoming.get("git_url")
    settings = Settings()
    project = Project(name=name, git_url=git_url)
    project.settings = settings
    db.session.add(settings)
    db.session.add(project)
    db.session.commit()

    return jsonify({"name": name, "git_url": git_url})



import psycopg2
import os
import contextlib
import sqlalchemy.pool as pool
import logging

from sqlalchemy import create_engine, MetaData
from sqlalchemy.orm import sessionmaker, scoped_session
from sqlalchemy.sql import text
from sqlalchemy.ext.declarative import declarative_base
from psycopg2.sql import Identifier, SQL


def engine_uri(**db_config):
    return "postgresql://{user}:{password}@{host}:{port}/{database}".format(**db_config)


SystemModel = declarative_base(metadata=MetaData(schema="meltano"))


class MetaDB(type):
    def __init__(cls, *_):
        cls._default = None

    @property
    def default(cls):
        return cls._default

    @default.setter
    def default(cls, v):
        if cls._default:
            cls._default.close()

        cls._default = v


class DB(metaclass=MetaDB):
    db_config = {
        "host": os.getenv("PG_ADDRESS", "localhost"),
        "port": os.getenv("PG_PORT", 5432),
        "user": os.getenv("PG_USERNAME", os.getenv("USER")),
        "password": os.getenv("PG_PASSWORD"),
        "database": os.getenv("PG_DATABASE"),
    }

    @classmethod
    def setup(cls, **kwargs):
        """
        Store the DB connection parameters and create a default engine.
        """
        cls.db_config.update(
            {k: kwargs[k] for k in cls.db_config.keys() if k in kwargs}
        )

        # use connection pooling for all connections
        pool.manage(psycopg2)
        cls.default = cls()

    @classmethod
    def create_database(cls, db_name):
        """
        Create a new database
        """
        config = cls.db_config.copy()
        config.update({"database": "postgres"})
        with create_engine(
            engine_uri(**config), isolation_level="AUTOCOMMIT"
        ).connect() as con:
            con.execute(text(f"CREATE DATABASE {db_name}"))

    @classmethod
    def close_default(cls):
        """
        Close the default engine
        """
        if cls.default:
            cls.default.close()

    @property
    def engine(self):
        return self._engine

    def __init__(self):
        self._engine = create_engine(engine_uri(**self.db_config))
        self._session_cls = scoped_session(sessionmaker(bind=self.engine))

    def create_connection(self):
        return psycopg2.connect(**self.db_config)

    def create_session(self):
        return self._session_cls()

    def open(self):
        return db_open(self)

    def session(self):
        return session_open(self)

    def close(self):
        if self.engine:
            self.engine.dispose()

    def ensure_schema_exists(self, schema_name, grant_roles=()):
        """
        Make sure that the given schema_name exists in the database
        If not, create it

        :param db_conn: psycopg2 database connection
        :param schema_name: database schema
        """
        conn = self.create_connection()
        cursor = conn.cursor()
        schema_identifier = Identifier(schema_name)
        group_identifiers = SQL(",").join(map(Identifier, grant_roles))

        create_schema = SQL(
            """
        CREATE SCHEMA IF NOT EXISTS {}
        """
        ).format(schema_identifier)

        grant_select_schema = SQL(
            """
        ALTER DEFAULT PRIVILEGES IN SCHEMA {} GRANT SELECT ON TABLES TO {}
        """
        ).format(schema_identifier, group_identifiers)

        grant_usage_schema = SQL(
            """
        GRANT USAGE ON SCHEMA {} TO {}
        """
        ).format(schema_identifier, group_identifiers)

        cursor.execute(create_schema)
        if grant_roles:
            cursor.execute(grant_select_schema)
            cursor.execute(grant_usage_schema)

        conn.commit()

        logging.info("Schema {} has been created successfully.".format(schema_name))
        for role in grant_roles:
            logging.info("Usage has been granted for role: {}.".format(role))


@contextlib.contextmanager
def db_open(db: DB = None):
    """Provide a raw connection in a transaction."""
    connection = (db or DB.default).create_connection()

    try:
        yield connection
        connection.commit()
    except Exception:
        connection.rollback()
        raise


@contextlib.contextmanager
def session_open(db: DB = None):
    """Provide a transactional scope around a series of operations."""
    session = (db or DB.default).create_session()
    try:
        yield session
        session.commit()
    except Exception:
        session.rollback()
        raise



import pytest
import yaml
import os
import shutil

from meltano.core.plugin_install_service import PluginInstallService
from meltano.core.config_service import ConfigService


class TestPluginInstallService:
    @pytest.fixture
    def subject(self, project):
        with project.meltanofile.open("w") as f:
            f.write(
                """
                    extractors:
                    - {name: tap-first, pip_url: 'git+https://gitlab.com/meltano/tap-first.git'}
                    loaders:
                    - {name: target-csv, pip_url: 'git+https://gitlab.com/meltano/target-csv.git'}
            """
            )

        return PluginInstallService(project)

    def test_default_init_should_not_fail(self, subject):
        assert subject

    @pytest.mark.slow
    def test_install_all(self, subject):
        all_plugins = subject.install_all_plugins()
        assert len(all_plugins["errors"]) == 0
        assert len(all_plugins["installed"]) == 2
        assert all_plugins["installed"][0]["status"] == "success"
        assert all_plugins["installed"][1]["status"] == "success"



from . import Job, State
from meltano.core.db import DB


class JobFinder:
    """
    Query builder for the `Job` model for a certain `elt_uri`.
    """

    def __init__(self, elt_uri):
        self.elt_uri = elt_uri

    def latest_success(self):
        with DB.default.session() as session:
            return (
                session.query(Job)
                .filter(
                    (Job.elt_uri == self.elt_uri)
                    & (Job.state == State.SUCCESS)
                    & Job.ended_at.isnot(None)
                )
                .order_by(Job.ended_at.desc())
                .first()
            )

    def latest_state(self):
        with DB.default.session() as session:
            return (
                session.query(Job)
                .filter(
                    (Job.elt_uri == self.elt_uri)
                    & Job.payload["singer_state"].isnot(None)
                    & Job.ended_at.isnot(None)
                )
                .order_by(Job.ended_at.desc())
                .first()
            )



import os
import subprocess
import logging
import re
from pathlib import Path

from meltano.core.project import Project
from meltano.core.plugin import Plugin


class PluginConfigService:
    """
    Manage the plugin configuration files. Each plugin can expose a
    set of files that should be stubbed using environment variables.
    """

    def __init__(self, project: Project, plugin: Plugin, config_dir=None, run_dir=None):
        self.project = project
        self.plugin = plugin

        # delegate to project
        self.config_dir = config_dir or project.plugin_dir(plugin)
        self.run_dir = run_dir or project.run_dir(plugin.name)

    @classmethod
    def envsubst(cls, src: Path, dst: Path, env={}):
        # find viable substitutions
        var_matcher = re.compile(
            """
            \$                 # starts with a '$'
            (?:                # either $VAR or ${VAR}
                {(\w+)}|(\w+)  # capture the variable name as group[0] or group[1]
            )
            """,
            re.VERBOSE,
        )

        env_override = os.environ.copy()
        env_override.update(env)

        def subst(match) -> str:
            try:
                # the variable can be in either group
                var = next(var for var in match.groups() if var)
                return str(env_override[var])
            except KeyError as e:
                logging.warning(f"Variable {var} is missing from the environment.")
                return None

        with src.open() as i, dst.open("w+") as o:
            output = re.sub(var_matcher, subst, i.read())
            o.write(output)

    def configure(self):
        os.makedirs(self.run_dir, exist_ok=True)

        config_file = self.config_dir.joinpath
        run_file = self.run_dir.joinpath

        # grab the list of files the plugin needs
        stubs = (
            (run_file(file_name), config_file(file_name))
            for file_name in self.plugin.config_files.values()
        )

        stubbed = []
        # stub the files from the env
        for dst, src in stubs:
            try:
                self.envsubst(src, dst)
                stubbed.append(dst)
            except FileNotFoundError:
                logging.warning(
                    f"Could not find {src.name} in {src.resolve()}, skipping."
                )

        return stubbed



import logging
import os
import json
import re
from typing import Dict
from functools import singledispatch

from . import Plugin, PluginType
from .error import TapDiscoveryError
from meltano.core.plugin_invoker import PluginInvoker
from meltano.core.utils import file_has_data
from meltano.core.behavior.hookable import HookObject, hook


class CatalogSelectAllVisitor:
    @singledispatch
    def visit(node, path: str = ""):
        logging.debug("Skipping node at '{path}'")

    @visit.register(dict)
    def _(node: dict, path=""):
        logging.debug("Visiting node at '{path}'.")
        if re.search(r"streams.\[\d+\]$", path):
            node.update({"selected": True})

        if path.endswith("metadata") and node.get("inclusion") == "available":
            node.update({"selected": True})
            logging.debug("{path} has been selected.")

        for child_path, child_node in node.items():
            CatalogSelectAllVisitor.visit(child_node, path=f"{path}.{child_path}")

    @visit.register(list)
    def _(node: list, path=""):
        logging.debug("Visiting node at '{path}'.")
        # ensure the stream metadata is in there
        if path.endswith("metadata"):
            for stream_metadata in (
                metadata for metadata in node if len(metadata["breadcrumb"]) == 0
            ):
                node["metadata"].update({"selected": True})
            else:
                node.append(
                    {
                        "breadcrumb": [],
                        "metadata": {"inclusion": "available", "selected": True},
                    }
                )
            logging.debug("{path} has been selected.")

        for index, child_node in enumerate(node):
            CatalogSelectAllVisitor.visit(child_node, path=f"{path}.[{index}]")


def plugin_factory(plugin_type: PluginType, canonical: Dict):
    plugin_class = {PluginType.EXTRACTORS: SingerTap, PluginType.LOADERS: SingerTarget}

    # this will parse the discovery file and create an instance of the
    # corresponding `plugin_class` for all the plugins.
    return plugin_class[plugin_type](**canonical)


class SingerPlugin(Plugin, HookObject):
    def __init__(self, *args, **kwargs):
        super().__init__(self.__class__.__plugin_type__, *args, **kwargs)

    @hook("before_install")
    def install_config_stub(self, project):
        plugin_dir = project.plugin_dir(self)
        os.makedirs(plugin_dir, exist_ok=True)

        # TODO: refactor as explicit stubs
        with open(plugin_dir.joinpath(self.config_files["config"]), "w") as config:
            json.dump(self.config, config)


class SingerTap(SingerPlugin):
    __plugin_type__ = PluginType.EXTRACTORS

    def exec_args(self, files: Dict):
        """
        Return the arguments list with the complete runtime paths.
        """
        args = ["--config", files["config"]]

        if file_has_data(files["catalog"]):
            args += ["--catalog", files["catalog"]]

        if file_has_data(files["state"]):
            args += ["--state", files["state"]]

        return args

    @property
    def config_files(self):
        return {
            "config": "tap.config.json",
            "catalog": "tap.properties.json",
            "state": "state.json",
        }

    @property
    def output_files(self):
        return {"output": "tap.out"}

    @hook("before_invoke")
    def run_discovery(self, plugin_invoker, exec_args):
        if not self._extras.get("autodiscover", True):
            return

        if "--discover" in exec_args:
            return

        properties_file = plugin_invoker.files["catalog"]

        with properties_file.open("w") as catalog:
            result = plugin_invoker.invoke("--discover", stdout=catalog)
            exit_code = result.wait()

        if exit_code != 0:
            logging.error(
                f"Command {plugin_invoker.exec_path()} {plugin_invoker.exec_args()} returned {exit_code}"
            )
            properties_file.unlink()
            return

        try:
            with properties_file.open() as catalog:
                schema = json.load(catalog)

            CatalogSelectAllVisitor.visit(schema)

            with properties_file.open("w") as catalog:
                json.dump(schema, catalog)
        except Exception as err:
            logging.exception(
                f"Could not select stream, catalog file is invalid: {properties_file}"
            )
            properties_file.unlink()


class SingerTarget(SingerPlugin):
    __plugin_type__ = PluginType.LOADERS

    def exec_args(self, files: Dict):
        args = ["--config", files["config"]]

        return args

    @property
    def config_files(self):
        return {"config": "target.config.json"}

    @property
    def output_files(self):
        return {"state": "new_state.json"}



import logging
import sqlalchemy.types as types

from datetime import datetime
from contextlib import contextmanager
from enum import Enum
from sqlalchemy import Column
from sqlalchemy.ext.mutable import MutableDict
from meltano.core.db import SystemModel, session_open
from meltano.core.error import Error


class InconsistentStateError(Error):
    """
    Occur upon a wrong operation for the current state.
    """


class ImpossibleTransitionError(Error):
    """
    Occur upon a wrong transition.
    """


class State(Enum):
    IDLE = (0, ("RUNNING", "FAIL"))
    RUNNING = (1, ("SUCCESS", "FAIL"))
    SUCCESS = (2, ())
    FAIL = (3, ("RUNNING",))
    DEAD = (4, ())

    def transitions(self):
        return self.value[1]

    def __str__(self):
        return self.name


class Job(SystemModel):
    __tablename__ = "job"

    id = Column(types.Integer, primary_key=True)
    elt_uri = Column(types.String)
    state = Column(types.Enum(State))
    started_at = Column(types.DateTime)
    ended_at = Column(types.DateTime)
    payload = Column(MutableDict.as_mutable(types.JSON))

    def __init__(self, **kwargs):
        kwargs["state"] = kwargs.get("state", State.IDLE)
        kwargs["payload"] = kwargs.get("payload", {})
        super().__init__(**kwargs)

    def can_transit(self, state: State) -> bool:
        if self.state is state:
            return True

        return state.name in self.state.transitions()

    def transit(self, state: State) -> (State, State):
        transition = (self.state, state)

        if not self.can_transit(state):
            raise ImpossibleTransitionError(transition)

        if self.state is state:
            return transition

        self.state = state
        self.save()

        return transition

    @contextmanager
    def run(self):
        try:
            self.start()
            yield
            self.success()
        except Exception as err:
            logging.error(err)
            self.fail(error=err)
            raise

    def start(self):
        self.started_at = datetime.utcnow()
        self.transit(State.RUNNING)
        self.save()

    def fail(self, error=None):
        self.ended_at = datetime.utcnow()
        self.transit(State.FAIL)
        self.payload = {"error": str(error)}
        self.save()

    def success(self):
        self.ended_at = datetime.utcnow()
        self.transit(State.SUCCESS)
        self.save()

    def __repr__(self):
        return "<Job(id='%s', elt_uri='%s', state='%s')>" % (
            self.id,
            self.elt_uri,
            self.state,
        )

    def save(self):
        with session_open() as session:
            session.add(self)
            return self



import base64
import json
import os
import subprocess
from os.path import join

import markdown
import pkg_resources
from flask import Blueprint, jsonify

reposBP = Blueprint("repos", __name__, url_prefix="/repos")

from ..app import db, meltano_model_path
from ..models.data import Model, Explore, View, Dimension, DimensionGroup, Measure, Join

path_to_parser = join(
    pkg_resources.resource_filename("meltano.api", "node_modules"),
    "lookml-parser/cli.js",
)
parser_command = [
    path_to_parser,
    "--input={}/*.{{view,model}}.lkml".format(meltano_model_path),
]


@reposBP.route("/", methods=["GET"])
def index():
    # rorepo is a Repo instance pointing to the git-python repository.
    # For all you know, the first argument to Repo is a path to the repository
    # you want to work with
    onlyfiles = [
        f
        for f in os.listdir(meltano_model_path)
        if os.path.isfile(os.path.join(meltano_model_path, f))
    ]
    sortedLkml = {"documents": [], "views": [], "models": [], "dashboards": []}
    for f in onlyfiles:
        filename, ext = os.path.splitext(f)
        file_dict = {"path": f, "abs": f, "visual": f}
        file_dict["unique"] = base64.b32encode(bytes(file_dict["abs"], "utf-8")).decode(
            "utf-8"
        )
        filename = filename.lower()
        if ext == ".md":
            sortedLkml["documents"].append(file_dict)
        else:
            filename, ext = os.path.splitext(filename)
            file_dict["visual"] = filename
            if ext == ".view":
                sortedLkml["views"].append(file_dict)
            if ext == ".model":
                sortedLkml["models"].append(file_dict)
            if ext == ".dashboard":
                sortedLkml["dashboards"].append(file_dict)

    return jsonify(sortedLkml)


@reposBP.route("/file/<unique>", methods=["GET"])
def file(unique):
    file_path = base64.b32decode(unique).decode("utf-8")
    (filename, ext) = os.path.splitext(file_path)
    is_markdown = False
    path_to_file = os.path.abspath(os.path.join(meltano_model_path, file_path))
    with open(path_to_file, "r") as read_file:
        data = read_file.read()
        if ext == ".md":
            data = markdown.markdown(data)
            is_markdown = True
        return jsonify(
            {
                "file": data,
                "is_markdown": is_markdown,
                "unique": unique,
                "populated": True,
            }
        )


@reposBP.route("/lint", methods=["GET"])
def lint():
    p = subprocess.run(parser_command, stdout=subprocess.PIPE)
    j = json.loads(p.stdout.decode("utf-8"))
    if "errors" in j:
        return jsonify({"result": False, "errors": j["errors"]})
    else:
        return jsonify({"result": True})


@reposBP.route("/update", methods=["GET"])
def db_import():
    p = subprocess.run(parser_command, stdout=subprocess.PIPE)
    j = json.loads(p.stdout.decode("utf-8"))

    # db.session.query(Explore).delete()
    # db.session.query(Model).delete()
    # db.session.query(View).delete()
    # db.session.query(Dimension).delete()
    # db.session.query(DimensionGroup).delete()
    # db.session.query(Measure).delete()
    # db.session.query(Join).delete()

    models = j["models"]
    files = j["files"]
    # process views
    for file in files:
        if file["_file_type"] == "view":
            file_view = file["views"][0]
            new_view_settings = {}
            if "label" in file_view:
                new_view_settings["label"] = file_view["label"].strip()
            if "sql_table_name" in file_view:
                new_view_settings["sql_table_name"] = file_view[
                    "sql_table_name"
                ].strip()
            new_view = View(file_view["_view"], new_view_settings)

            # Add dimensions for view
            if "dimensions" in file_view:
                for dimension in file_view["dimensions"]:
                    new_dimension_settings = {}
                    if "hidden" in dimension:
                        new_dimension_settings["hidden"] = dimension["hidden"]
                    if "primary_key" in dimension:
                        new_dimension_settings["primary_key"] = dimension["primary_key"]
                    if "label" in dimension:
                        new_dimension_settings["label"] = dimension["label"].strip()
                    if "type" in dimension:
                        new_dimension_settings["type"] = dimension["type"].strip()
                    if "sql" in dimension:
                        new_dimension_settings["sql"] = dimension["sql"].strip()
                    new_dimension_settings["_type"] = dimension["_type"]
                    new_dimension_settings["_n"] = dimension["_n"]
                    new_dimension = Dimension(
                        dimension["_dimension"], new_dimension_settings
                    )
                    new_view.dimensions.append(new_dimension)
                    db.session.add(new_dimension)
                    new_view.dimensions.append(new_dimension)

            if "dimension_groups" in file_view:
                for dimension_group in file_view["dimension_groups"]:
                    new_dimension_group_settings = {}
                    if "timeframes" in dimension_group:
                        new_dimension_group_settings["timeframes"] = dimension_group[
                            "timeframes"
                        ]
                    if "sql" in dimension_group:
                        new_dimension_group_settings["sql"] = dimension_group["sql"]
                    if "type" in dimension_group:
                        new_dimension_group_settings["type"] = dimension_group["type"]
                    if "label" in dimension_group:
                        new_dimension_group_settings["label"] = dimension_group["label"]
                    if "convert_tz" in dimension_group:
                        new_dimension_group_settings["convert_tz"] = dimension_group[
                            "convert_tz"
                        ]
                    if "description" in dimension_group:
                        new_dimension_group_settings["description"] = dimension_group[
                            "description"
                        ]

                    new_dimension_group_settings["_n"] = dimension_group["_n"]
                    new_dimension_group_settings["_type"] = dimension_group["_type"]
                    new_dimension_group = DimensionGroup(
                        dimension_group["_dimension_group"],
                        new_dimension_group_settings,
                    )
                    new_view.dimension_groups.append(new_dimension_group)
                    db.session.add(new_dimension_group)

            # Add measures for view
            if "measures" in file_view:
                for measure in file_view["measures"]:
                    new_measure_settings = {}
                    if "hidden" in measure:
                        new_measure_settings["hidden"] = measure["hidden"]
                    if "value_format" in measure:
                        new_measure_settings["value_format"] = measure[
                            "value_format"
                        ].strip()
                    if "label" in measure:
                        new_measure_settings["label"] = measure["label"].strip()
                    if "type" in measure:
                        new_measure_settings["type"] = measure["type"].strip()
                    if "sql" in measure:
                        new_measure_settings["sql"] = measure["sql"].strip()
                    new_measure_settings["_type"] = measure["_type"].strip()
                    new_measure_settings["_n"] = measure["_n"]
                    new_measure = Measure(measure["_measure"], new_measure_settings)
                    new_view.measures.append(new_measure)
                    db.session.add(new_measure)
                    new_view.measures.append(new_measure)
            db.session.add(new_view)
    # process models
    for model in models:
        model_settings = {}
        if "label" in model:
            model_settings["label"] = model["label"]
        model_settings["include"] = model["include"]
        model_settings["connection"] = model["connection"]
        model_settings["_type"] = model["_type"]

        new_model = Model(model["_model"], model_settings)

        # Set the explores for the model
        if len(model["explores"]):
            for explore in model["explores"]:
                explore_settings = {}
                if "label" in explore:
                    explore_settings["label"] = explore["label"]
                if "view_label" in explore:
                    explore_settings["view_label"] = explore["view_label"]
                if "description" in explore:
                    explore_settings["description"] = explore["description"]
                if "always_filter" in explore:
                    explore_settings["always_filter"] = explore["always_filter"]
                explore_settings["_type"] = explore["_type"]

                new_explore = Explore(explore["_explore"], explore_settings)

                # Set the view for the explore
                # Name the explore from `from` or from name of explore itself
                view_name = explore.get("from", explore["_explore"])
                connected_view = View.query.filter_by(name=view_name).first()
                new_explore.view = connected_view
                if "joins" in explore:
                    explore_joins = explore["joins"]
                    for join in explore_joins:
                        explore_join_settings = {}
                        if "view_label" in join:
                            explore_join_settings["view_label"] = join["view_label"]
                        if "label" in join:
                            explore_join_settings["label"] = join["label"]
                        if "type" in join:
                            explore_join_settings["type"] = join["type"]
                        if "relationship" in join:
                            explore_join_settings["relationship"] = join["relationship"]
                        if "sql_on" in join:
                            explore_join_settings["sql_on"] = join["sql_on"]
                        if "type" in join:
                            explore_join_settings["type"] = join["type"]

                        new_explore_join = Join(join["_join"], explore_join_settings)
                        new_explore.joins.append(new_explore_join)
                        db.session.add(new_explore_join)

                new_model.explores.append(new_explore)
                db.session.add(new_explore)

        db.session.add(new_model)
    # project = Project(name=name, git_url=git_url)
    db.session.commit()
    return jsonify({"result": True})


@reposBP.route("/test", methods=["GET"])
def db_test():
    explore = Explore.query.first()
    return jsonify({"explore": {"name": explore.name, "settings": explore.settings}})


@reposBP.route("/models", methods=["GET"])
def models():
    models = Model.query.all()
    models_json = []
    for model in models:
        this_model = {}
        this_model["settings"] = model.settings
        this_model["name"] = model.name
        this_model["explores"] = []
        for explore in model.explores:
            this_explore = {}
            this_explore["settings"] = explore.settings
            this_explore["name"] = explore.name
            this_explore["link"] = f"/explore/{model.name}/{explore.name}"
            this_view = {}
            this_view["name"] = explore.view.name
            this_view["settings"] = explore.view.settings
            this_explore["views"] = this_view
            this_explore["joins"] = []
            for join in explore.joins:
                this_join = {}
                this_join["name"] = join.name
                this_join["settings"] = join.settings
                this_explore["joins"].append(this_join)

            this_model["explores"].append(this_explore)
        models_json.append(this_model)
    return jsonify(models_json)


@reposBP.route("/explores", methods=["GET"])
def explores():
    explores = Explore.query.all()
    explores_json = []
    for explore in explores:
        explores_json.append(explore.serializable())
    return jsonify(explores_json)


@reposBP.route("/views/<view_name>", methods=["GET"])
def view_read(view_name):
    view = View.query.filter(View.name == view_name).first()
    return jsonify(view.serializable(True))


@reposBP.route("/explores/<model_name>/<explore_name>", methods=["GET"])
def explore_read(model_name, explore_name):
    explore = (
        Explore.query.join(Model, Explore.model_id == Model.id)
        .filter(Model.name == model_name)
        .filter(Explore.name == explore_name)
        .first()
    )
    explore_json = explore.serializable(True)
    explore_json["settings"]["has_filters"] = False
    if "always_filter" in explore_json["settings"]:
        explore_json["settings"]["has_filters"] = True
        for a_filter in explore_json["settings"]["always_filter"]["filters"]:
            dimensions = explore_json["view"]["dimensions"]
            for dimension in dimensions:
                if dimension["name"] == a_filter["field"]:
                    a_filter["explore_label"] = a_filter["_explore"].title()
                    a_filter["type"] = dimension["settings"]["type"]
                    if "label" in dimension["settings"]:
                        a_filter["label"] = dimension["settings"]["label"]
                    else:
                        a_filter["label"] = " ".join(
                            dimension["name"].split("_")
                        ).title()
                    a_filter["sql"] = dimension["settings"]["sql"]
                    break
    return jsonify(explore_json)



import pytest
import json
from unittest import mock
from pathlib import Path

from meltano.core.plugin import PluginType
from meltano.core.plugin_invoker import PluginInvoker
from meltano.core.plugin.singer import CatalogSelectAllVisitor


CATALOG = """
{
  "streams": [
    {
      "tap_stream_id": "Account",
      "stream": "Account",
      "key_properties": [
        "Id"
      ],
      "schema": {
        "type": "object",
        "additionalProperties": false,
        "properties": {
          "AccountCode__c": {
            "type": [
              "string",
              "null"
            ]
          },
          "AccountNumber": {
            "type": [
              "string",
              "null"
            ]
          },
          "AdditionalEmailAddresses": {
            "type": [
              "string",
              "null"
            ]
          },
          "AllowInvoiceEdit": {
            "type": [
              "boolean",
              "null"
            ]
          },
          "AutoPay": {
            "type": [
              "boolean",
              "null"
            ]
          },
          "Balance": {
            "type": [
              "number",
              "null"
            ]
          },
          "Batch": {
            "type": [
              "string",
              "null"
            ]
          },
          "BcdSettingOption": {
            "type": [
              "string",
              "null"
            ]
          },
          "BillCycleDay": {
            "type": [
              "integer",
              "null"
            ]
          },
          "CommunicationProfileId": {
            "type": [
              "string",
              "null"
            ]
          },
          "ConversionRate__c": {
            "type": [
              "string",
              "null"
            ]
          },
          "CreatedById": {
            "type": [
              "string",
              "null"
            ]
          },
          "CreatedDate": {
            "type": [
              "string",
              "null"
            ],
            "format": "date-time"
          },
          "CreditBalance": {
            "type": [
              "number",
              "null"
            ]
          },
          "CrmId": {
            "type": [
              "string",
              "null"
            ]
          },
          "Currency": {
            "type": [
              "string",
              "null"
            ]
          },
          "CustomerServiceRepName": {
            "type": [
              "string",
              "null"
            ]
          },
          "Entity__c": {
            "type": [
              "string",
              "null"
            ]
          },
          "Id": {
            "type": [
              "string",
              "null"
            ]
          },
          "InvoiceDeliveryPrefsEmail": {
            "type": [
              "boolean",
              "null"
            ]
          },
          "InvoiceDeliveryPrefsPrint": {
            "type": [
              "boolean",
              "null"
            ]
          },
          "InvoiceTemplateId": {
            "type": [
              "string",
              "null"
            ]
          },
          "LastInvoiceDate": {
            "type": [
              "string",
              "null"
            ],
            "format": "date-time"
          },
          "Mrr": {
            "type": [
              "number",
              "null"
            ]
          },
          "Name": {
            "type": [
              "string",
              "null"
            ]
          },
          "Notes": {
            "type": [
              "string",
              "null"
            ]
          },
          "Parent__c": {
            "type": [
              "string",
              "null"
            ]
          },
          "ParentId": {
            "type": [
              "string",
              "null"
            ]
          },
          "PaymentGateway": {
            "type": [
              "string",
              "null"
            ]
          },
          "PaymentTerm": {
            "type": [
              "string",
              "null"
            ]
          },
          "PurchaseOrderNumber": {
            "type": [
              "string",
              "null"
            ]
          },
          "SalesRepName": {
            "type": [
              "string",
              "null"
            ]
          },
          "SFDC_Sync__c": {
            "type": [
              "string",
              "null"
            ]
          },
          "Status": {
            "type": [
              "string",
              "null"
            ]
          },
          "Support_Hold__c__c": {
            "type": [
              "string",
              "null"
            ]
          },
          "TaxCompanyCode": {
            "type": [
              "string",
              "null"
            ]
          },
          "TaxExemptCertificateID": {
            "type": [
              "string",
              "null"
            ]
          },
          "TaxExemptCertificateType": {
            "type": [
              "string",
              "null"
            ]
          },
          "TaxExemptDescription": {
            "type": [
              "string",
              "null"
            ]
          },
          "TaxExemptEffectiveDate": {
            "type": [
              "string",
              "null"
            ],
            "format": "date-time"
          },
          "TaxExemptEntityUseCode": {
            "type": [
              "string",
              "null"
            ]
          },
          "TaxExemptExpirationDate": {
            "type": [
              "string",
              "null"
            ],
            "format": "date-time"
          },
          "TaxExemptIssuingJurisdiction": {
            "type": [
              "string",
              "null"
            ]
          },
          "TaxExemptStatus": {
            "type": [
              "string",
              "null"
            ]
          },
          "TotalDebitMemoBalance": {
            "type": [
              "number",
              "null"
            ]
          },
          "TotalInvoiceBalance": {
            "type": [
              "number",
              "null"
            ]
          },
          "UnappliedBalance": {
            "type": [
              "number",
              "null"
            ]
          },
          "UnappliedCreditMemoAmount": {
            "type": [
              "number",
              "null"
            ]
          },
          "UpdatedById": {
            "type": [
              "string",
              "null"
            ]
          },
          "UpdatedDate": {
            "type": [
              "string",
              "null"
            ],
            "format": "date-time"
          },
          "VATId": {
            "type": [
              "string",
              "null"
            ]
          },
          "BillToContactId": {
            "type": [
              "string",
              "null"
            ]
          },
          "DefaultPaymentMethodId": {
            "type": [
              "string",
              "null"
            ]
          },
          "ParentAccountId": {
            "type": [
              "string",
              "null"
            ]
          },
          "SoldToContactId": {
            "type": [
              "string",
              "null"
            ]
          },
          "Deleted": {
            "type": "boolean"
          }
        }
      },
      "metadata": [
        {
          "breadcrumb": [
            "properties",
            "AccountCode__c"
          ],
          "metadata": {
              "inclusion": "available"
          }
        },
        {
          "breadcrumb": [
            "properties",
            "AccountNumber"
          ],
          "metadata": {
            "inclusion": "available"
          }
        },
        {
          "breadcrumb": [
            "properties",
            "AdditionalEmailAddresses"
          ],
          "metadata": {
            "inclusion": "available"
          }
        },
        {
          "breadcrumb": [
            "properties",
            "AllowInvoiceEdit"
          ],
          "metadata": {
            "inclusion": "available"
          }
        },
        {
          "breadcrumb": [
            "properties",
            "AutoPay"
          ],
          "metadata": {
            "inclusion": "available"
          }
        },
        {
          "breadcrumb": [
            "properties",
            "Balance"
          ],
          "metadata": {
            "inclusion": "available"
          }
        },
        {
          "breadcrumb": [
            "properties",
            "Batch"
          ],
          "metadata": {
            "inclusion": "available"
          }
        },
        {
          "breadcrumb": [
            "properties",
            "BcdSettingOption"
          ],
          "metadata": {
            "inclusion": "available"
          }
        },
        {
          "breadcrumb": [
            "properties",
            "BillCycleDay"
          ],
          "metadata": {
            "inclusion": "available"
          }
        },
        {
          "breadcrumb": [
            "properties",
            "CommunicationProfileId"
          ],
          "metadata": {
            "inclusion": "available"
          }
        },
        {
          "breadcrumb": [
            "properties",
            "ConversionRate__c"
          ],
          "metadata": {
            "inclusion": "available"
          }
        },
        {
          "breadcrumb": [
            "properties",
            "CreatedById"
          ],
          "metadata": {
            "inclusion": "available"
          }
        },
        {
          "breadcrumb": [
            "properties",
            "CreatedDate"
          ],
          "metadata": {
            "inclusion": "available"
          }
        },
        {
          "breadcrumb": [
            "properties",
            "CreditBalance"
          ],
          "metadata": {
            "inclusion": "available"
          }
        },
        {
          "breadcrumb": [
            "properties",
            "CrmId"
          ],
          "metadata": {
            "inclusion": "available"
          }
        },
        {
          "breadcrumb": [
            "properties",
            "Currency"
          ],
          "metadata": {
            "inclusion": "available"
          }
        },
        {
          "breadcrumb": [
            "properties",
            "CustomerServiceRepName"
          ],
          "metadata": {
            "inclusion": "available"
          }
        },
        {
          "breadcrumb": [
            "properties",
            "Entity__c"
          ],
          "metadata": {
            "inclusion": "available"
          }
        },
        {
          "breadcrumb": [
            "properties",
            "Id"
          ],
          "metadata": {
            "inclusion": "available"
          }
        },
        {
          "breadcrumb": [
            "properties",
            "InvoiceDeliveryPrefsEmail"
          ],
          "metadata": {
            "inclusion": "available"
          }
        },
        {
          "breadcrumb": [
            "properties",
            "InvoiceDeliveryPrefsPrint"
          ],
          "metadata": {
            "inclusion": "available"
          }
        },
        {
          "breadcrumb": [
            "properties",
            "InvoiceTemplateId"
          ],
          "metadata": {
            "inclusion": "available"
          }
        },
        {
          "breadcrumb": [
            "properties",
            "LastInvoiceDate"
          ],
          "metadata": {
            "inclusion": "available"
          }
        },
        {
          "breadcrumb": [
            "properties",
            "Mrr"
          ],
          "metadata": {
            "inclusion": "available"
          }
        },
        {
          "breadcrumb": [
            "properties",
            "Name"
          ],
          "metadata": {
            "inclusion": "available"
          }
        },
        {
          "breadcrumb": [
            "properties",
            "Notes"
          ],
          "metadata": {
            "inclusion": "available"
          }
        },
        {
          "breadcrumb": [
            "properties",
            "Parent__c"
          ],
          "metadata": {
            "inclusion": "available"
          }
        },
        {
          "breadcrumb": [
            "properties",
            "ParentId"
          ],
          "metadata": {
            "inclusion": "available"
          }
        },
        {
          "breadcrumb": [
            "properties",
            "PaymentGateway"
          ],
          "metadata": {
            "inclusion": "available"
          }
        },
        {
          "breadcrumb": [
            "properties",
            "PaymentTerm"
          ],
          "metadata": {
            "inclusion": "available"
          }
        },
        {
          "breadcrumb": [
            "properties",
            "PurchaseOrderNumber"
          ],
          "metadata": {
            "inclusion": "available"
          }
        },
        {
          "breadcrumb": [
            "properties",
            "SalesRepName"
          ],
          "metadata": {
            "inclusion": "available"
          }
        },
        {
          "breadcrumb": [
            "properties",
            "SFDC_Sync__c"
          ],
          "metadata": {
            "inclusion": "available"
          }
        },
        {
          "breadcrumb": [
            "properties",
            "Status"
          ],
          "metadata": {
            "inclusion": "available"
          }
        },
        {
          "breadcrumb": [
            "properties",
            "Support_Hold__c__c"
          ],
          "metadata": {
            "inclusion": "available"
          }
        },
        {
          "breadcrumb": [
            "properties",
            "TaxCompanyCode"
          ],
          "metadata": {
            "inclusion": "available"
          }
        },
        {
          "breadcrumb": [
            "properties",
            "TaxExemptCertificateID"
          ],
          "metadata": {
            "inclusion": "available"
          }
        },
        {
          "breadcrumb": [
            "properties",
            "TaxExemptCertificateType"
          ],
          "metadata": {
            "inclusion": "available"
          }
        },
        {
          "breadcrumb": [
            "properties",
            "TaxExemptDescription"
          ],
          "metadata": {
            "inclusion": "available"
          }
        },
        {
          "breadcrumb": [
            "properties",
            "TaxExemptEffectiveDate"
          ],
          "metadata": {
            "inclusion": "available"
          }
        },
        {
          "breadcrumb": [
            "properties",
            "TaxExemptEntityUseCode"
          ],
          "metadata": {
            "inclusion": "available"
          }
        },
        {
          "breadcrumb": [
            "properties",
            "TaxExemptExpirationDate"
          ],
          "metadata": {
            "inclusion": "available"
          }
        },
        {
          "breadcrumb": [
            "properties",
            "TaxExemptIssuingJurisdiction"
          ],
          "metadata": {
            "inclusion": "available"
          }
        },
        {
          "breadcrumb": [
            "properties",
            "TaxExemptStatus"
          ],
          "metadata": {
            "inclusion": "available"
          }
        },
        {
          "breadcrumb": [
            "properties",
            "TotalDebitMemoBalance"
          ],
          "metadata": {
            "inclusion": "available"
          }
        },
        {
          "breadcrumb": [
            "properties",
            "TotalInvoiceBalance"
          ],
          "metadata": {
            "inclusion": "available"
          }
        },
        {
          "breadcrumb": [
            "properties",
            "UnappliedBalance"
          ],
          "metadata": {
            "inclusion": "available"
          }
        },
        {
          "breadcrumb": [
            "properties",
            "UnappliedCreditMemoAmount"
          ],
          "metadata": {
            "inclusion": "available"
          }
        },
        {
          "breadcrumb": [
            "properties",
            "UpdatedById"
          ],
          "metadata": {
            "inclusion": "available"
          }
        },
        {
          "breadcrumb": [
            "properties",
            "UpdatedDate"
          ],
          "metadata": {
            "inclusion": "available"
          }
        },
        {
          "breadcrumb": [
            "properties",
            "VATId"
          ],
          "metadata": {
            "inclusion": "available"
          }
        },
        {
          "breadcrumb": [
            "properties",
            "BillToContactId"
          ],
          "metadata": {
            "tap-zuora.joined_object": "BillToContact",
            "inclusion": "available"
          }
        },
        {
          "breadcrumb": [
            "properties",
            "DefaultPaymentMethodId"
          ],
          "metadata": {
            "tap-zuora.joined_object": "DefaultPaymentMethod",
            "inclusion": "available"
          }
        },
        {
          "breadcrumb": [
            "properties",
            "ParentAccountId"
          ],
          "metadata": {
            "tap-zuora.joined_object": "ParentAccount",
            "inclusion": "available"
          }
        },
        {
          "breadcrumb": [
            "properties",
            "SoldToContactId"
          ],
          "metadata": {
            "tap-zuora.joined_object": "SoldToContact",
            "inclusion": "available"
          }
        },
        {
          "breadcrumb": [
            "properties",
            "Deleted"
          ],
          "metadata": {
            "inclusion": "available"
          }
        }
      ],
      "replication_key": "UpdatedDate",
      "replication_method": "INCREMENTAL"
    },
    {
      "tap_stream_id": "AccountingCode",
      "stream": "AccountingCode",
      "key_properties": [
        "Id"
      ],
      "schema": {
        "type": "object",
        "additionalProperties": false,
        "properties": {
          "Category": {
            "type": [
              "string",
              "null"
            ]
          },
          "CreatedById": {
            "type": [
              "string",
              "null"
            ]
          },
          "CreatedDate": {
            "type": [
              "string",
              "null"
            ],
            "format": "date-time"
          },
          "GLAccountName": {
            "type": [
              "string",
              "null"
            ]
          },
          "GLAccountNumber": {
            "type": [
              "string",
              "null"
            ]
          },
          "Id": {
            "type": [
              "string",
              "null"
            ]
          },
          "Name": {
            "type": [
              "string",
              "null"
            ]
          },
          "Notes": {
            "type": [
              "string",
              "null"
            ]
          },
          "Status": {
            "type": [
              "string",
              "null"
            ]
          },
          "Type": {
            "type": [
              "string",
              "null"
            ]
          },
          "UpdatedById": {
            "type": [
              "string",
              "null"
            ]
          },
          "UpdatedDate": {
            "type": [
              "string",
              "null"
            ],
            "format": "date-time"
          },
          "Deleted": {
            "type": "boolean"
          }
        }
      },
      "metadata": [
        {
          "breadcrumb": [
            "properties",
            "Category"
          ],
          "metadata": {
            "inclusion": "available"
          }
        },
        {
          "breadcrumb": [
            "properties",
            "CreatedById"
          ],
          "metadata": {
            "inclusion": "available"
          }
        },
        {
          "breadcrumb": [
            "properties",
            "CreatedDate"
          ],
          "metadata": {
            "inclusion": "available"
          }
        },
        {
          "breadcrumb": [
            "properties",
            "GLAccountName"
          ],
          "metadata": {
            "inclusion": "available"
          }
        },
        {
          "breadcrumb": [
            "properties",
            "GLAccountNumber"
          ],
          "metadata": {
            "inclusion": "available"
          }
        },
        {
          "breadcrumb": [
            "properties",
            "Id"
          ],
          "metadata": {
            "inclusion": "available"
          }
        },
        {
          "breadcrumb": [
            "properties",
            "Name"
          ],
          "metadata": {
            "inclusion": "automatic"
          }
        },
        {
          "breadcrumb": [
            "properties",
            "Notes"
          ],
          "metadata": {
            "inclusion": "available"
          }
        },
        {
          "breadcrumb": [
            "properties",
            "Status"
          ],
          "metadata": {
            "inclusion": "available"
          }
        },
        {
          "breadcrumb": [
            "properties",
            "Type"
          ],
          "metadata": {
            "inclusion": "available"
          }
        },
        {
          "breadcrumb": [
            "properties",
            "UpdatedById"
          ],
          "metadata": {
            "inclusion": "available"
          }
        },
        {
          "breadcrumb": [
            "properties",
            "UpdatedDate"
          ],
          "metadata": {
            "inclusion": "available"
          }
        },
        {
          "breadcrumb": [
            "properties",
            "Deleted"
          ],
          "metadata": {
            "inclusion": "available"
          }
        }
      ],
      "replication_key": "UpdatedDate",
      "replication_method": "INCREMENTAL"
    }
  ]
}
"""


class TestSingerTap:
    @pytest.fixture
    def subject(self, project_add_service):
        return project_add_service.add(PluginType.EXTRACTORS, "tap-first")

    def config_files(self, subject, dir: Path):
        return {key: dir.join(file) for key, file in subject.config_files.items()}

    def test_exec_args(self, subject, tmpdir):
        base_files = self.config_files(subject, tmpdir.mkdir("base"))
        assert subject.exec_args(base_files) == ["--config", base_files["config"]]

        # when `catalog` has data
        base_files = self.config_files(subject, tmpdir.mkdir("catalog"))
        base_files["catalog"].open("w").write("...")
        assert subject.exec_args(base_files) == [
            "--config",
            base_files["config"],
            "--catalog",
            base_files["catalog"],
        ]

        # when `state` has data
        base_files = self.config_files(subject, tmpdir.mkdir("state"))
        base_files["state"].open("w").write("...")
        assert subject.exec_args(base_files) == [
            "--config",
            base_files["config"],
            "--state",
            base_files["state"],
        ]

    def test_run_discovery(self, project, subject):
        process_mock = mock.Mock()
        process_mock.wait.return_value = 0

        invoker = PluginInvoker(project, subject)
        invoker.prepare()

        with mock.patch.object(
            PluginInvoker, "invoke", return_value=process_mock
        ) as invoke:
            subject.run_discovery(invoker, [])

            assert invoke.called_with(["--discover"])

    def test_run_discovery_fails(self, project, subject):
        process_mock = mock.Mock()
        process_mock.wait.return_value = 1  # something wrong happened

        invoker = PluginInvoker(project, subject)
        invoker.prepare()

        with mock.patch.object(
            PluginInvoker, "invoke", return_value=process_mock
        ) as invoke:
            subject.run_discovery(invoker, [])

            assert not invoker.files[
                "catalog"
            ].exists(), "Catalog should not be present."

    def test_run_discovery_invalid(self, project, subject):
        process_mock = mock.Mock()
        process_mock.wait.return_value = 0

        invoker = PluginInvoker(project, subject)
        invoker.prepare()

        def corrupt_catalog(*_, **__):
            invoker.files["catalog"].open("w").write("this is invalid json")

            return process_mock

        with mock.patch.object(
            PluginInvoker, "invoke", side_effect=corrupt_catalog
        ) as invoke:
            subject.run_discovery(invoker, [])

            assert not invoker.files[
                "catalog"
            ].exists(), "Catalog should not be present."


class TestCatalogSelectVisitor:
    @pytest.fixture
    def catalog(self):
        return json.loads(CATALOG)

    @classmethod
    def stream_is_selected(cls, stream):
        return stream.get("selected", False)

    @classmethod
    def metadata_is_selected(cls, metadata):
        inclusion = metadata.get("inclusion")
        if inclusion == "automatic":
            return true

        if inclusion == "available":
            return metadata.get("selected", False)

    def test_visit(self, catalog):
        CatalogSelectAllVisitor.visit(catalog)

        streams = {stream["stream"]: stream for stream in catalog["streams"]}

        metadatas = {
            stream["stream"]: metadata
            for _, stream in streams.items()
            for metadata in stream["metadata"]
        }

        # all streams are selected
        for name, stream in streams.items():
            assert self.stream_is_selected(stream), f"{name} is not selected."

        # all fields are selecte
        for stream, metadata in metadatas.items():
            field_metadata = metadata["metadata"]
            assert self.metadata_is_selected(
                field_metadata
            ), f"{stream}.{metadata['breadcrumb']} is not selected"


class TestSingerTarget:
    @pytest.fixture
    def subject(self, project_add_service):
        return project_add_service.add(PluginType.LOADERS, "target-csv")

    def test_exec_args(self, subject):
        base_files = subject.config_files
        assert subject.exec_args(base_files) == ["--config", base_files["config"]]
